# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from collections import defaultdict
import os
import datetime
import logging
import time
import struct
from enum import Enum
from spinn_utilities.overrides import overrides
from spinn_utilities.log import FormatAdapter
from spinnman.exceptions import SpinnmanTimeoutException
from spinnman.messages.sdp import SDPMessage, SDPHeader, SDPFlag
from spinnman.messages.scp.impl.iptag_set import IPTagSet
from spinnman.connections.udp_packet_connections import SCAMPConnection
from spinnman.model.enums.cpu_state import CPUState
from pacman.executor.injection_decorator import inject_items
from pacman.model.graphs.common import EdgeTrafficType
from pacman.model.graphs.machine import MachineVertex
from pacman.model.resources import (
    ConstantSDRAM, IPtagResource, ResourceContainer)
from spinn_front_end_common.utilities.globals_variables import get_simulator
from spinn_front_end_common.utilities.helpful_functions import (
    convert_vertices_to_core_subset, n_word_struct)
from spinn_front_end_common.utilities.emergency_recovery import (
    emergency_recover_state_from_failure)
from spinn_front_end_common.abstract_models import (
    AbstractHasAssociatedBinary, AbstractGeneratesDataSpecification)
from spinn_front_end_common.interface.provenance import (
    AbstractProvidesLocalProvenanceData)
from spinn_front_end_common.utilities.utility_objs import (
    ExecutableType, ProvenanceDataItem)
from spinn_front_end_common.utilities.constants import (
    SDP_PORTS, BYTES_PER_WORD, BYTES_PER_KB)
from spinn_front_end_common.utilities.exceptions import SpinnFrontEndException
from spinn_front_end_common.utilities.utility_objs.\
    extra_monitor_scp_processes import (
        SetRouterTimeoutProcess, ClearQueueProcess)

log = FormatAdapter(logging.getLogger(__name__))

# shift by for the destination x coord in the word.
DEST_X_SHIFT = 16

TIMEOUT_RETRY_LIMIT = 100
TIMEOUT_MESSAGE = "Failed to hear from the machine during {} attempts. "\
    "Please try removing firewalls."
_MINOR_LOSS_MESSAGE = (
    "During the extraction of data of {} bytes from memory address {}, "
    "attempt {} had {} sequences that were lost.")
_MINOR_LOSS_THRESHOLD = 10
_MAJOR_LOSS_MESSAGE = (
    "During the extraction of data from chip {}, there were {} cases of "
    "serious data loss. The system recovered, but the speed of download "
    "was compromised. Reduce the number of executing applications and remove "
    "routers between yourself and the SpiNNaker machine to reduce the chance "
    "of this occurring.")
_MAJOR_LOSS_THRESHOLD = 100

# cap for stopping wrap arounds
TRANSACTION_ID_CAP = 0xFFFFFFFF

#: number of items used up by the retransmit code for its header
SDP_RETRANSMISSION_HEADER_SIZE = 2

#: size of config region in bytes
#: 1.new seq key, 2.first data key, 3. transaction id key 4.end flag key,
# 5.base key, 6.iptag tag
CONFIG_SIZE = 6 * BYTES_PER_WORD

#: items of data a SDP packet can hold when SCP header removed
WORDS_PER_FULL_PACKET = 68  # 272 bytes as removed SCP header

#: size of items the sequence number uses
SEQUENCE_NUMBER_SIZE_IN_ITEMS = 1

#: transaction id size in words
TRANSACTION_ID_SIZE_IN_ITEMS = 1

#: the size in words of the command flag
COMMAND_SIZE_IN_ITEMS = 1

#: offset for missing seq starts in first packet
WORDS_FOR_COMMAND_N_MISSING_TRANSACTION = 3

#: offset for missing seq starts in more packet
WORDS_FOR_COMMAND_TRANSACTION = (
    COMMAND_SIZE_IN_ITEMS + TRANSACTION_ID_SIZE_IN_ITEMS)

BYTES_FOR_SEQ_AND_TRANSACTION_ID = (
    (SEQUENCE_NUMBER_SIZE_IN_ITEMS + TRANSACTION_ID_SIZE_IN_ITEMS) *
    BYTES_PER_WORD)

#: items of data from SDP packet with a sequence number
WORDS_PER_FULL_PACKET_WITH_SEQUENCE_NUM = (
    WORDS_PER_FULL_PACKET - SEQUENCE_NUMBER_SIZE_IN_ITEMS -
    TRANSACTION_ID_SIZE_IN_ITEMS)

#: offset where data in starts on commands
#: (command, transaction_id, seq num)
WORDS_FOR_COMMAND_AND_KEY = 3
BYTES_FOR_COMMAND_AND_ADDRESS_HEADER = (
    WORDS_FOR_COMMAND_AND_KEY * BYTES_PER_WORD)

#: offset where data in starts in reception (command, transaction id)
WORDS_FOR_RECEPTION_COMMAND_AND_ADDRESS_HEADER = 2
BYTES_FOR_RECEPTION_COMMAND_AND_ADDRESS_HEADER = (
    WORDS_FOR_RECEPTION_COMMAND_AND_ADDRESS_HEADER * BYTES_PER_WORD)

#: size for data to store when first packet with command and address
WORDS_IN_FULL_PACKET_WITH_KEY = (
    WORDS_PER_FULL_PACKET - WORDS_FOR_COMMAND_AND_KEY)
BYTES_IN_FULL_PACKET_WITH_KEY = (
    WORDS_IN_FULL_PACKET_WITH_KEY * BYTES_PER_WORD)

#: size of data in key space
#: x, y, key (all ints) for possible 48 chips, plus n chips to read,
# the reinjector base key.
SIZE_DATA_IN_CHIP_TO_KEY_SPACE = ((3 * 48) + 2) * BYTES_PER_WORD


class _DATA_REGIONS(Enum):
    """DSG data regions"""
    CONFIG = 0
    CHIP_TO_KEY_SPACE = 1


class DATA_OUT_COMMANDS(Enum):
    """command IDs for the SDP packets for data out"""
    START_SENDING = 100
    START_MISSING_SEQ = 1000
    MISSING_SEQ = 1001
    CLEAR = 2000


class DATA_IN_COMMANDS(Enum):
    """command IDs for the SDP packets for data in"""
    SEND_DATA_TO_LOCATION = 200
    SEND_SEQ_DATA = 2000
    SEND_TELL = 2001
    RECEIVE_MISSING_SEQ_DATA = 2002
    RECEIVE_FINISHED = 2003


# precompiled structures
_ONE_WORD = struct.Struct("<I")
_TWO_WORDS = struct.Struct("<II")
_THREE_WORDS = struct.Struct("<III")
_FOUR_WORDS = struct.Struct("<IIII")
_FIVE_WORDS = struct.Struct("<IIIII")

# Set to true to check that the data is correct after it has been sent in.
# This is expensive, and only works in Python 3.5 or later.
VERIFY_SENT_DATA = False


def ceildiv(dividend, divisor):
    """ How to divide two possibly-integer numbers and round up.
    """
    assert divisor > 0
    q, r = divmod(dividend, divisor)
    return int(q) + (r != 0)


# SDRAM requirement for storing missing SDP packets seq nums
SDRAM_FOR_MISSING_SDP_SEQ_NUMS = ceildiv(
    120.0 * 1024 * BYTES_PER_KB,
    WORDS_PER_FULL_PACKET_WITH_SEQUENCE_NUM * BYTES_PER_WORD)


class DataSpeedUpPacketGatherMachineVertex(
        MachineVertex, AbstractGeneratesDataSpecification,
        AbstractHasAssociatedBinary, AbstractProvidesLocalProvenanceData):
    """ Machine vertex for handling fast data transfer between host and \
        SpiNNaker. This machine vertex is only ever placed on chips with a \
        working Ethernet connection; it collaborates with the \
        :py:class:`ExtraMonitorSupportMachineVertex` to write data on other \
        chips..
    """
    __slots__ = [
        # x coordinate
        "_x",
        # y coordinate
        "_y",
        # word with x and y
        "_coord_word",
        # transaction id
        "_transaction_id",
        # app id
        "_app_id",
        # socket
        "_connection",
        # store of the extra monitors to location. helpful in data in
        "_extra_monitors_by_chip",
        # path for the data in report
        "_in_report_path",
        # ipaddress
        "_ip_address",
        # store for the last reinjection status
        "_last_status",
        # the max seq num expected given a data retrieval
        "_max_seq_num",
        # holder for missing seq nums for data in
        "_missing_seq_nums_data_in",
        # holder of data from out
        "_output",
        # my placement for future lookup
        "_placement",
        # provenance holder
        "_provenance_data_items",
        # Count of the runs for provenance data
        "_run",
        "_remote_tag",
        # path to the data out report
        "_out_report_path",
        # bool flag for writing reports
        "_write_data_speed_up_reports",
        # data holder for output
        "_view"]

    #: base key (really nasty hack to tie in fixed route keys)
    BASE_KEY = 0xFFFFFFF9
    NEW_SEQ_KEY = 0xFFFFFFF8
    FIRST_DATA_KEY = 0xFFFFFFF7
    END_FLAG_KEY = 0xFFFFFFF6
    TRANSACTION_ID_KEY = 0xFFFFFFF5

    #: to use with multicast stuff (reinjection acks have to be fixed route)
    BASE_MASK = 0xFFFFFFFB
    NEW_SEQ_KEY_OFFSET = 1
    FIRST_DATA_KEY_OFFSET = 2
    END_FLAG_KEY_OFFSET = 3
    TRANSACTION_ID_KEY_OFFSET = 4

    # throttle on the transmission
    _TRANSMISSION_THROTTLE_TIME = 0.000001

    # TRAFFIC_TYPE = EdgeTrafficType.MULTICAST
    TRAFFIC_TYPE = EdgeTrafficType.FIXED_ROUTE

    #: report name for tracking used routers
    OUT_REPORT_NAME = "routers_used_in_speed_up_process.rpt"
    #: report name for tracking performance gains
    IN_REPORT_NAME = "speeds_gained_in_speed_up_process.rpt"

    # the end flag is set when the high bit of the sequence number word is set
    _LAST_MESSAGE_FLAG_BIT_MASK = 0x80000000
    # corresponding mask for the actual sequence numbers
    _SEQUENCE_NUMBER_MASK = 0x7fffffff

    # time outs used by the protocol for separate bits
    _TIMEOUT_PER_RECEIVE_IN_SECONDS = 2
    _TIMEOUT_FOR_SENDING_IN_SECONDS = 0.01

    # end flag for missing seq nums
    _MISSING_SEQ_NUMS_END_FLAG = 0xFFFFFFFF

    # flag for saying missing all SEQ numbers
    FLAG_FOR_MISSING_ALL_SEQUENCES = 0xFFFFFFFE

    _ADDRESS_PACKET_BYTE_FORMAT = struct.Struct(
        "<{}B".format(BYTES_IN_FULL_PACKET_WITH_KEY))

    # Router timeouts, in mantissa,exponent form. See datasheet for details
    _LONG_TIMEOUT = (14, 14)
    _SHORT_TIMEOUT = (1, 1)
    _TEMP_TIMEOUT = (15, 4)
    _ZERO_TIMEOUT = (0, 0)

    # Initial port for the reverse IP tag (to be replaced later)
    _TAG_INITIAL_PORT = 10000

    def __init__(
            self, x, y, extra_monitors_by_chip, ip_address,
            report_default_directory,
            write_data_speed_up_reports, app_vertex=None, constraints=None):
        """
        :param int x: Where this gatherer is.
        :param int y: Where this gatherer is.
        :param extra_monitors_by_chip: UNUSED
        :type extra_monitors_by_chip:
            dict(tuple(int,int), ExtraMonitorSupportMachineVertex)
        :param str ip_address:
            How to talk directly to the chip where the gatherer is.
        :param str report_default_directory: Where reporting is done.
        :param bool write_data_speed_up_reports:
            Whether to write low-level reports on data transfer speeds.
        :param constraints:
        :type constraints:
            iterable(~pacman.model.constraints.AbstractConstraint) or None
        :param app_vertex:
            The application vertex that caused this machine vertex to be
            created. If None, there is no such application vertex.
        :type app_vertex:
            ~pacman.model.graphs.application.ApplicationVertex or None
        """
        super().__init__(
            label="SYSTEM:PacketGatherer({},{})".format(x, y),
            constraints=constraints, app_vertex=app_vertex)

        # data holders for the output, and sequence numbers
        self._view = None
        self._max_seq_num = None
        self._output = None
        self._transaction_id = 0

        # store of the extra monitors to location. helpful in data in
        self._extra_monitors_by_chip = extra_monitors_by_chip
        self._missing_seq_nums_data_in = list()

        # Create a connection to be used
        self._x = x
        self._y = y
        self._coord_word = None
        self._ip_address = ip_address
        self._remote_tag = None
        self._connection = None

        # local provenance storage
        self._provenance_data_items = defaultdict(list)
        self._run = 0
        self._placement = None
        self._app_id = None

        # create report if it doesn't already exist
        self._out_report_path = \
            os.path.join(report_default_directory, self.OUT_REPORT_NAME)
        self._in_report_path = \
            os.path.join(report_default_directory, self.IN_REPORT_NAME)
        self._write_data_speed_up_reports = write_data_speed_up_reports

        # Stored reinjection status for resetting timeouts
        self._last_status = None

    def __throttled_send(self, message):
        """ slows down transmissions to allow spinnaker to keep up.

        :param ~.SDPMessage message: message to send
        :rtype: None
        """
        # send first message
        self._connection.send_sdp_message(message)
        time.sleep(self._TRANSMISSION_THROTTLE_TIME)

    @property
    @overrides(MachineVertex.resources_required)
    def resources_required(self):
        return self.static_resources_required()

    def update_transaction_id_from_machine(self, txrx):
        """ Looks up from the machine what the current transaction ID is\
            and updates the data speed up gatherer.

        :param ~spinnman.transceiver.Transceiver txrx: SpiNNMan instance
        """
        self._transaction_id = txrx.read_user_1(
            self._placement.x, self._placement.y, self._placement.p)

    @classmethod
    def static_resources_required(cls):
        """
        :rtype: ~pacman.model.resources.ResourceContainer
        """
        return ResourceContainer(
            sdram=ConstantSDRAM(
                CONFIG_SIZE + SDRAM_FOR_MISSING_SDP_SEQ_NUMS +
                SIZE_DATA_IN_CHIP_TO_KEY_SPACE),
            iptags=[IPtagResource(
                port=cls._TAG_INITIAL_PORT, strip_sdp=True,
                ip_address="localhost", traffic_identifier="DATA_SPEED_UP")])

    @overrides(AbstractHasAssociatedBinary.get_binary_start_type)
    def get_binary_start_type(self):
        return ExecutableType.SYSTEM

    @inject_items({
        "machine_graph": "MemoryMachineGraph",
        "routing_info": "MemoryRoutingInfos",
        "tags": "MemoryTags",
        "mc_data_chips_to_keys": "DataInMulticastKeyToChipMap",
        "machine": "MemoryExtendedMachine",
        "app_id": "APPID",
        "router_timeout_key": "SystemMulticastRouterTimeoutKeys"
    })
    @overrides(
        AbstractGeneratesDataSpecification.generate_data_specification,
        additional_arguments={
            "machine_graph", "routing_info", "tags",
            "mc_data_chips_to_keys", "machine", "app_id",
            "router_timeout_key"
        })
    def generate_data_specification(
            self, spec, placement, machine_graph, routing_info, tags,
            mc_data_chips_to_keys, machine, app_id, router_timeout_key):
        """
        :param ~pacman.model.graphs.machine.MachineGraph machine_graph:
            (injected)
        :param ~pacman.model.routing_info.RoutingInfo routing_info: (injected)
        :param ~pacman.model.tags.Tags tags: (injected)
        :param dict(tuple(int,int),int) mc_data_chips_to_keys: (injected)
        :param ~spinn_machine.Machine machine: (injected)
        :param int app_id: (injected)
        :param dict(tuple(int,int),int) router_timeout_key: (injected)
        """
        # pylint: disable=too-many-arguments, arguments-differ

        # update my placement for future knowledge
        self._placement = placement
        self._app_id = app_id

        # Create the data regions for hello world
        self._reserve_memory_regions(spec)

        # the keys for the special cases
        if self.TRAFFIC_TYPE == EdgeTrafficType.MULTICAST:
            base_key = routing_info.get_first_key_for_edge(
                list(machine_graph.get_edges_ending_at_vertex(self))[0])
            new_seq_key = base_key + self.NEW_SEQ_KEY_OFFSET
            first_data_key = base_key + self.FIRST_DATA_KEY_OFFSET
            end_flag_key = base_key + self.END_FLAG_KEY_OFFSET
            transaction_id_key = base_key + self.TRANSACTION_ID_KEY_OFFSET
        else:
            new_seq_key = self.NEW_SEQ_KEY
            first_data_key = self.FIRST_DATA_KEY
            end_flag_key = self.END_FLAG_KEY
            base_key = self.BASE_KEY
            transaction_id_key = self.TRANSACTION_ID_KEY

        spec.switch_write_focus(_DATA_REGIONS.CONFIG.value)
        spec.write_value(new_seq_key)
        spec.write_value(first_data_key)
        spec.write_value(transaction_id_key)
        spec.write_value(end_flag_key)
        spec.write_value(base_key)

        # locate the tag ID for our data and update with a port
        # Note: The port doesn't matter as we are going to override this later
        iptags = tags.get_ip_tags_for_vertex(self)
        iptag = iptags[0]
        spec.write_value(iptag.tag)
        self._remote_tag = iptag.tag

        # write mc chip key map
        spec.switch_write_focus(_DATA_REGIONS.CHIP_TO_KEY_SPACE.value)
        chips_on_board = list(machine.get_existing_xys_on_board(
            machine.get_chip_at(placement.x, placement.y)))

        # write how many chips to read
        spec.write_value(len(chips_on_board))

        # write the broad cast keys for timeouts
        reinjection_base_key = router_timeout_key[(placement.x, placement.y)]
        spec.write_value(reinjection_base_key)

        # write each chip x and y and base key
        for chip_xy in chips_on_board:
            board_chip_x, board_chip_y = machine.get_local_xy(
                machine.get_chip_at(*chip_xy))
            spec.write_value(board_chip_x)
            spec.write_value(board_chip_y)
            spec.write_value(mc_data_chips_to_keys[chip_xy])
            log.debug("for chip {}:{} base key is {}",
                      chip_xy[0], chip_xy[1], mc_data_chips_to_keys[chip_xy])

        # End-of-Spec:
        spec.end_specification()

    @staticmethod
    def _reserve_memory_regions(spec):
        """ Writes the DSG regions memory sizes. Static so that it can be used\
            by the application vertex.

        :param ~.DataSpecificationGenerator spec: spec file
        """
        spec.reserve_memory_region(
            region=_DATA_REGIONS.CONFIG.value,
            size=CONFIG_SIZE,
            label="config")
        spec.reserve_memory_region(
            region=_DATA_REGIONS.CHIP_TO_KEY_SPACE.value,
            size=SIZE_DATA_IN_CHIP_TO_KEY_SPACE,
            label="mc_key_map")

    @overrides(AbstractHasAssociatedBinary.get_binary_file_name)
    def get_binary_file_name(self):
        return "data_speed_up_packet_gatherer.aplx"

    @overrides(AbstractProvidesLocalProvenanceData.get_local_provenance_data)
    def get_local_provenance_data(self):
        self._run += 1
        prov_items = list()
        significant_losses = defaultdict(list)
        top_level_name = "Provenance_for_{}".format(self._label)
        for (placement, memory_address, length_in_bytes) in \
                self._provenance_data_items.keys():

            # handle duplicates of the same calls
            times_extracted_the_same_thing = 0
            for time_taken, lost_seq_nums in self._provenance_data_items[
                    placement, memory_address, length_in_bytes]:
                # handle time
                chip_name = "chip{}:{}".format(placement.x, placement.y)
                last_name = "Memory_address:{}:Length_in_bytes:{}"\
                    .format(memory_address, length_in_bytes)
                if times_extracted_the_same_thing == 0:
                    iteration_name = "run{}".format(
                        self._run)
                else:
                    iteration_name = "run{}iteration{}".format(
                        self._run, times_extracted_the_same_thing)
                prov_items.append(ProvenanceDataItem(
                    [top_level_name, "extraction_time", chip_name, last_name,
                     iteration_name],
                    time_taken, report=False))
                times_extracted_the_same_thing += 1

                # handle lost sequence numbers
                for i, n_lost_seq_nums in enumerate(lost_seq_nums):
                    # Zeroes are not reported at all
                    if n_lost_seq_nums:
                        prov_items.append(ProvenanceDataItem(
                            [top_level_name, "lost_seq_nums", chip_name,
                             last_name, iteration_name,
                             "iteration_{}".format(i)],
                            n_lost_seq_nums, report=(
                                n_lost_seq_nums > _MINOR_LOSS_THRESHOLD),
                            message=_MINOR_LOSS_MESSAGE.format(
                                length_in_bytes, memory_address, i,
                                n_lost_seq_nums)))
                    if n_lost_seq_nums > _MAJOR_LOSS_THRESHOLD:
                        significant_losses[placement.x, placement.y] += [i]
        for chip in significant_losses:
            n_times = len(significant_losses[chip])
            chip_name = "chip{}:{}".format(*chip)
            prov_items.append(ProvenanceDataItem(
                [top_level_name, "serious_lost_seq_num_count", chip_name],
                n_times, report=True, message=_MAJOR_LOSS_MESSAGE.format(
                    chip, n_times)))
        self._provenance_data_items = defaultdict(list)
        return prov_items

    @staticmethod
    def locate_correct_write_data_function_for_chip_location(
            uses_advanced_monitors, machine, x, y, transceiver,
            extra_monitor_cores_to_ethernet_connection_map):
        """ Supports other components figuring out which gatherer and function\
            to call for writing data onto SpiNNaker.

        :param bool uses_advanced_monitors:
            Whether the system is using advanced monitors
        :param ~spinn_machine.Machine machine: the SpiNNMachine instance
        :param int x: the chip x coordinate to write data to
        :param int y: the chip y coordinate to write data to
        :param ~spinnman.transceiver.Transceiver transceiver:
            the SpiNNMan instance
        :param extra_monitor_cores_to_ethernet_connection_map:
            mapping between cores and connections
        :type extra_monitor_cores_to_ethernet_connection_map:
            dict(tuple(int,int), DataSpeedUpPacketGatherMachineVertex)
        :return: a write function of either a LPG or the spinnMan
        :rtype: callable
        """
        if not uses_advanced_monitors:
            return transceiver.write_memory

        chip = machine.get_chip_at(x, y)
        ethernet_connected_chip = machine.get_chip_at(
            chip.nearest_ethernet_x, chip.nearest_ethernet_y)
        gatherer = extra_monitor_cores_to_ethernet_connection_map[
            ethernet_connected_chip.x, ethernet_connected_chip.y]
        return gatherer.send_data_into_spinnaker

    def _generate_data_in_report(
            self, time_diff, data_size, x, y,
            address_written_to, missing_seq_nums):
        """ Writes the data in report for this stage.

        :param ~datetime.timedelta time_diff:
            the time taken to write the memory
        :param int data_size: the size of data that was written in bytes
        :param int x:
            the location in machine where the data was written to X axis
        :param int y:
            the location in machine where the data was written to Y axis
        :param int address_written_to: where in SDRAM it was written to
        :param list(set(int)) missing_seq_nums:
            the set of missing sequence numbers per data transmission attempt
        :rtype: None
        """
        if not os.path.isfile(self._in_report_path):
            with open(self._in_report_path, "w") as writer:
                writer.write(
                    "x\t\t y\t\t SDRAM address\t\t size in bytes\t\t\t"
                    " time took \t\t\t Mb/s \t\t\t missing sequence numbers\n")
                writer.write(
                    "------------------------------------------------"
                    "------------------------------------------------"
                    "-------------------------------------------------\n")

        time_took_ms = float(time_diff.microseconds +
                             time_diff.total_seconds() * 1000000)
        megabits = (data_size * 8.0) / (1024 * BYTES_PER_KB)
        if time_took_ms == 0:
            mbs = "unknown, below threshold"
        else:
            mbs = megabits / (float(time_took_ms) / 100000.0)

        with open(self._in_report_path, "a") as writer:
            writer.write(
                "{}\t\t {}\t\t {}\t\t {}\t\t\t\t {}\t\t\t {}\t\t {}\n".format(
                    x, y, address_written_to, data_size, time_took_ms,
                    mbs, missing_seq_nums))

    def send_data_into_spinnaker(
            self, x, y, base_address, data, n_bytes=None, offset=0,
            cpu=0, is_filename=False):  # pylint: disable=unused-argument
        """ Sends a block of data into SpiNNaker to a given chip.

        :param int x: chip x for data
        :param int y: chip y for data
        :param int base_address: the address in SDRAM to start writing memory
        :param data: the data to write (or filename to load data from,
            if ``is_filename`` is True; that's the only time this is a str)
        :type data: bytes or bytearray or memoryview or str
        :param int n_bytes: how many bytes to read, or None if not set
        :param int offset: where in the data to start from
        :param int cpu:
        :param bool is_filename: whether data is actually a file.
        """
        # if file, read in and then process as normal
        if is_filename:
            if offset != 0:
                raise Exception(
                    "when using a file, you can only have a offset of 0")

            with open(data, "rb") as reader:
                # n_bytes=None already means 'read everything'
                data = reader.read(n_bytes)  # pylint: disable=no-member
            # Number of bytes to write is now length of buffer we have
            n_bytes = len(data)
        elif n_bytes is None:
            n_bytes = len(data)
        transceiver = get_simulator().transceiver

        # start time recording
        start = datetime.datetime.now()
        # send data
        self._send_data_via_extra_monitors(
            transceiver, x, y, base_address, data[offset:n_bytes + offset])
        # end time recording
        end = datetime.datetime.now()

        if VERIFY_SENT_DATA:
            original_data = bytes(data[offset:n_bytes + offset])
            verified_data = bytes(transceiver.read_memory(
                x, y, base_address, n_bytes))
            self.__verify_sent_data(
                original_data, verified_data, x, y, base_address, n_bytes)

        # write report
        if self._write_data_speed_up_reports:
            self._generate_data_in_report(
                x=x, y=y, time_diff=end - start,
                data_size=n_bytes, address_written_to=base_address,
                missing_seq_nums=self._missing_seq_nums_data_in)

    @staticmethod
    def __verify_sent_data(
            original_data, verified_data, x, y, base_address, n_bytes):
        if original_data != verified_data:
            log.error("VARIANCE: chip:{},{} address:{} len:{}",
                      x, y, base_address, n_bytes)
            log.error("original:{}", original_data.hex())
            log.error("verified:{}", verified_data.hex())
            for i, (a, b) in enumerate(zip(original_data, verified_data)):
                if a != b:
                    raise Exception("damn at " + str(i))

    @staticmethod
    def __make_sdp_message(placement, port, payload):
        """
        :param ~.Placement placement:
        :param SDP_PORTS port:
        :param bytearray payload:
        :rtype: ~.SDPMessage
        """
        return SDPMessage(
            sdp_header=SDPHeader(
                destination_chip_x=placement.x,
                destination_chip_y=placement.y,
                destination_cpu=placement.p,
                destination_port=port.value,
                flags=SDPFlag.REPLY_NOT_EXPECTED),
            data=payload)

    def _send_data_via_extra_monitors(
            self, transceiver, destination_chip_x, destination_chip_y,
            start_address, data_to_write):
        """ sends data using the extra monitor cores

        :param ~.Transceiver transceiver: the SpiNNMan instance
        :param int destination_chip_x: chip x
        :param int destination_chip_y: chip y
        :param int start_address: start address in SDRAM to write data to
        :param bytearray data_to_write: the data to write
        :param int start_address: the base SDRAM address
        """
        # Set up the connection
        self._connection = SCAMPConnection(
            chip_x=self._x, chip_y=self._y, remote_host=self._ip_address)
        self.__reprogram_tag(self._connection)

        # how many packets after first one we need to send
        self._max_seq_num = ceildiv(
            len(data_to_write), BYTES_IN_FULL_PACKET_WITH_KEY)

        # determine board chip IDs, as the LPG does not know machine scope IDs
        machine = transceiver.get_machine_details()
        chip = machine.get_chip_at(destination_chip_x, destination_chip_y)
        dest_x, dest_y = machine.get_local_xy(chip)
        self._coord_word = (dest_x << DEST_X_SHIFT) | dest_y

        # for safety, check the transaction id from the machine before updating
        self.update_transaction_id_from_machine(transceiver)
        self._transaction_id = (self._transaction_id + 1) & TRANSACTION_ID_CAP
        time_out_count = 0

        # verify completed
        received_confirmation = False
        while not received_confirmation:

            # send initial attempt at sending all the data
            self._send_all_data_based_packets(data_to_write, start_address)

            # Don't create a missing buffer until at least one packet has
            # come back.
            missing = None

            while not received_confirmation:
                try:
                    # try to receive a confirmation of some sort from spinnaker
                    data = self._connection.receive(
                        timeout=self._TIMEOUT_PER_RECEIVE_IN_SECONDS)
                    time_out_count = 0

                    # Read command and transaction id
                    (cmd, transaction_id) = _TWO_WORDS.unpack_from(data, 0)

                    # If wrong transaction id, ignore packet
                    if self._transaction_id != transaction_id:
                        continue

                    # Decide what to do with the packet
                    if cmd == DATA_IN_COMMANDS.RECEIVE_FINISHED.value:
                        received_confirmation = True
                        break

                    if cmd != DATA_IN_COMMANDS.RECEIVE_MISSING_SEQ_DATA.value:
                        raise Exception("Unknown command {} received".format(
                            cmd))

                    # The currently received packet has missing sequence
                    # numbers. Accumulate and dispatch transactionId when
                    # we've got them all.
                    if missing is None:
                        missing = set()
                        self._missing_seq_nums_data_in.append(missing)
                    seen_last, seen_all = self._read_in_missing_seq_nums(
                        data, BYTES_FOR_RECEPTION_COMMAND_AND_ADDRESS_HEADER,
                        missing)

                    # Check that you've seen something that implies ready
                    # to retransmit.
                    if seen_all or seen_last:
                        self._outgoing_retransmit_missing_seq_nums(
                            data_to_write, missing)
                        missing.clear()

                except SpinnmanTimeoutException as e:
                    # if the timeout has not occurred x times, keep trying
                    time_out_count += 1
                    if time_out_count > TIMEOUT_RETRY_LIMIT:
                        emergency_recover_state_from_failure(
                            transceiver, self._app_id, self, self._placement)
                        raise SpinnFrontEndException(
                            TIMEOUT_MESSAGE.format(
                                time_out_count)) from e

                    # If we never received a packet, we will never have
                    # created the buffer, so send everything again
                    if missing is None:
                        break

                    self._outgoing_retransmit_missing_seq_nums(
                            data_to_write, missing)
                    missing.clear()

        # Close the connection
        self._connection.close()

    def _read_in_missing_seq_nums(self, data, position, seq_nums):
        """ handles a missing seq num packet from spinnaker

        :param data: the data to translate into missing seq nums
        :type data: bytearray or bytes
        :param int position: the position in the data to write.
        :param set(int) seq_nums: a set of sequence numbers to add to
        :return: seen_last flag and seen_all flag
        :rtype: tuple(bool, bool)
        """
        # find how many elements are in this packet
        n_elements = (len(data) - position) // BYTES_PER_WORD

        # store missing
        new_seq_nums = n_word_struct(n_elements).unpack_from(
            data, position)

        # add missing seqs accordingly
        seen_last = False
        seen_all = False
        if new_seq_nums[-1] == self._MISSING_SEQ_NUMS_END_FLAG:
            new_seq_nums = new_seq_nums[:-1]
            seen_last = True
        if new_seq_nums[-1] == self.FLAG_FOR_MISSING_ALL_SEQUENCES:
            for missing_seq in range(0, self._max_seq_num):
                seq_nums.add(missing_seq)
            seen_all = True
        else:
            seq_nums.update(new_seq_nums)

        return seen_last, seen_all

    def _outgoing_retransmit_missing_seq_nums(
            self, data_to_write, missing):
        """ Transmits back into SpiNNaker the missing data based off missing\
            sequence numbers

        :param bytearray data_to_write: the data to write.
        :param set(int) missing: a set of missing sequence numbers
        """

        missing_seqs_as_list = list(missing)
        missing_seqs_as_list.sort()

        # send seq data
        for missing_seq_num in missing_seqs_as_list:
            message, _length = self._calculate_data_in_data_from_seq_number(
                data_to_write, missing_seq_num,
                DATA_IN_COMMANDS.SEND_SEQ_DATA.value, None)
            self.__throttled_send(message)

        # request an update on what is missing
        self._send_tell_flag()

    @staticmethod
    def _calculate_position_from_seq_number(seq_num):
        """ Calculates where in the raw data to start reading from, given a\
            sequence number

        :param int seq_num: the sequence number to determine position from
        :return: the position in the byte data
        :rtype: int
        """
        return BYTES_IN_FULL_PACKET_WITH_KEY * seq_num

    def _calculate_data_in_data_from_seq_number(
            self, data_to_write, seq_num, command_id, position):
        """ Determine the data needed to be sent to the SpiNNaker machine\
            given a sequence number

        :param bytearray data_to_write:
            the data to write to the SpiNNaker machine
        :param int seq_num: the seq num to get the data for
        :param int command_id:
        :param position: the position in the data to write to spinnaker
        :type position: int or None
        :return: SDP message and how much data has been written
        :rtype: tuple(~.SDPMessage, int)
        """

        # check for last packet
        packet_data_length = BYTES_IN_FULL_PACKET_WITH_KEY

        # determine position in data if not given
        if position is None:
            position = self._calculate_position_from_seq_number(seq_num)

        # if less than a full packet worth of data, adjust length
        if position + packet_data_length > len(data_to_write):
            packet_data_length = len(data_to_write) - position

        if packet_data_length < 0:
            raise Exception("weird packet data length.")

        # determine the true packet length (with header)
        packet_length = (
            packet_data_length + BYTES_FOR_COMMAND_AND_ADDRESS_HEADER)

        # create struct
        packet_data = bytearray(packet_length)
        _THREE_WORDS.pack_into(
            packet_data, 0, command_id, self._transaction_id, seq_num)
        struct.pack_into(
            "<{}B".format(packet_data_length), packet_data,
            BYTES_FOR_COMMAND_AND_ADDRESS_HEADER,
            *data_to_write[position:position+packet_data_length])

        # debug
        # self._print_out_packet_data(packet_data, position)

        # build sdp packet
        message = self.__make_sdp_message(
            self._placement, SDP_PORTS.EXTRA_MONITOR_CORE_DATA_IN_SPEED_UP,
            packet_data)

        # return message for sending, and the length in data sent
        return message, packet_data_length

    def _send_location(self, start_address):
        """ Send location as separate message.

        :param int start_address: SDRAM location
        """
        self._connection.send_sdp_message(self.__make_sdp_message(
            self._placement, SDP_PORTS.EXTRA_MONITOR_CORE_DATA_IN_SPEED_UP,
            _FIVE_WORDS.pack(
                DATA_IN_COMMANDS.SEND_DATA_TO_LOCATION.value,
                self._transaction_id, start_address, self._coord_word,
                self._max_seq_num - 1)))
        log.debug(
            "start address for transaction {} is {}".format(
                self._transaction_id, start_address))

    def _send_tell_flag(self):
        """ Send tell flag as separate message.
        """
        self._connection.send_sdp_message(self.__make_sdp_message(
            self._placement, SDP_PORTS.EXTRA_MONITOR_CORE_DATA_IN_SPEED_UP,
            _TWO_WORDS.pack(
                DATA_IN_COMMANDS.SEND_TELL.value, self._transaction_id)))

    def _send_all_data_based_packets(self, data_to_write, start_address):
        """ Send all the data as one block.

        :param bytearray data_to_write: the data to send
        :param int start_address:
        """
        # Send the location
        self._send_location(start_address)

        # where in the data we are currently up to
        position_in_data = 0

        # send rest of data
        for seq_num in range(0, self._max_seq_num):
            # put in command flag and seq num
            message, length_to_send = \
                self._calculate_data_in_data_from_seq_number(
                    data_to_write, seq_num,
                    DATA_IN_COMMANDS.SEND_SEQ_DATA.value, position_in_data)
            position_in_data += length_to_send

            # send the message
            self.__throttled_send(message)
            log.debug("sent seq {} of {} bytes", seq_num, length_to_send)

        # check for end flag
        self._send_tell_flag()
        log.debug("sent end flag")

    @staticmethod
    def streaming(gatherers, transceiver, extra_monitor_cores, placements):
        """ Helper method for setting the router timeouts to a state usable\
            for data streaming via a Python context manager (i.e., using\
            the `with` statement).

        :param list(DataSpeedUpPacketGatherMachineVertex) gatherers:
            All the gatherers that are to be set
        :param ~spinnman.transceiver.Transceiver transceiver:
            the SpiNNMan instance
        :param list(ExtraMonitorSupportMachineVertex) extra_monitor_cores:
            the extra monitor cores to set
        :param ~pacman.model.placements.Placements placements:
            placements object
        :return: a context manager
        """
        return _StreamingContextManager(
            gatherers, transceiver, extra_monitor_cores, placements)

    def set_cores_for_data_streaming(
            self, transceiver, extra_monitor_cores, placements):
        """ Helper method for setting the router timeouts to a state usable\
            for data streaming.

        :param ~spinnman.transceiver.Transceiver transceiver:
            the SpiNNMan instance
        :param list(ExtraMonitorSupportMachineVertex) extra_monitor_cores:
            the extra monitor cores to set
        :param ~pacman.model.placements.Placements placements:
            placements object
        """
        lead_monitor = extra_monitor_cores[0]
        # Store the last reinjection status for resetting
        # NOTE: This assumes the status is the same on all cores
        self._last_status = lead_monitor.get_reinjection_status(
            placements, transceiver)

        # Set to not inject dropped packets
        lead_monitor.set_reinjection_packets(
            placements, extra_monitor_cores, transceiver,
            point_to_point=False, multicast=False, nearest_neighbour=False,
            fixed_route=False)

        # Clear any outstanding packets from reinjection
        self.clear_reinjection_queue(transceiver, placements)

        # set time outs
        self.set_router_wait2_timeout(
            self._SHORT_TIMEOUT, transceiver, placements)
        self.set_router_wait1_timeout(
            self._LONG_TIMEOUT, transceiver, placements)

    @staticmethod
    def load_application_routing_tables(
            transceiver, extra_monitor_cores, placements):
        """ Set all chips to have application table loaded in the router.

        :param ~spinnman.transceiver.Transceiver transceiver:
            the SpiNNMan instance
        :param list(ExtraMonitorSupportMachineVertex) extra_monitor_cores:
            the extra monitor cores to set
        :param ~pacman.model.placements.Placements placements:
            placements object
        """
        extra_monitor_cores[0].load_application_mc_routes(
            placements, extra_monitor_cores, transceiver)

    @staticmethod
    def load_system_routing_tables(
            transceiver, extra_monitor_cores, placements):
        """ Set all chips to have the system table loaded in the router

        :param ~spinnman.transceiver.Transceiver transceiver:
            the SpiNNMan instance
        :param list(ExtraMonitorSupportMachineVertex) extra_monitor_cores:
            the extra monitor cores to set
        :param ~pacman.model.placements.Placements placements:
            placements object
        """
        extra_monitor_cores[0].load_system_mc_routes(
            placements, extra_monitor_cores, transceiver)

    def set_router_wait1_timeout(self, timeout, transceiver, placements):
        """ Set the wait1 field for a set of routers.

        :param tuple(int,int) timeout:
        :param ~spinnman.transceiver.Transceiver transceiver:
        :param ~pacman.model.placements.Placements placements:
        """
        mantissa, exponent = timeout
        core_subsets = convert_vertices_to_core_subset([self], placements)
        process = SetRouterTimeoutProcess(
            transceiver.scamp_connection_selector)
        try:
            process.set_wait1_timeout(mantissa, exponent, core_subsets)
        except:  # noqa: E722
            emergency_recover_state_from_failure(
                transceiver, self._app_id, self,
                placements.get_placement_of_vertex(self))
            raise

    def set_router_wait2_timeout(self, timeout, transceiver, placements):
        """ Set the wait2 field for a set of routers.

        :param tuple(int,int) timeout:
        :param ~spinnman.transceiver.Transceiver transceiver:
        :param ~pacman.model.placements.Placements placements:
        """
        mantissa, exponent = timeout
        core_subsets = convert_vertices_to_core_subset([self], placements)
        process = SetRouterTimeoutProcess(
            transceiver.scamp_connection_selector)
        try:
            process.set_wait2_timeout(mantissa, exponent, core_subsets)
        except:  # noqa: E722
            emergency_recover_state_from_failure(
                transceiver, self._app_id, self,
                placements.get_placement_of_vertex(self))
            raise

    def clear_reinjection_queue(self, transceiver, placements):
        """ Clears the queues for reinjection.

        :param ~spinnman.transceiver.Transceiver transceiver:
            the spinnMan interface
        :param ~pacman.model.placements.Placements placements:
            the placements object
        """
        core_subsets = convert_vertices_to_core_subset([self], placements)
        process = ClearQueueProcess(transceiver.scamp_connection_selector)
        try:
            process.reset_counters(core_subsets)
        except:  # noqa: E722
            emergency_recover_state_from_failure(
                transceiver, self._app_id, self,
                placements.get_placement_of_vertex(self))
            raise

    def unset_cores_for_data_streaming(
            self, transceiver, extra_monitor_cores, placements):
        """ Helper method for restoring the router timeouts to normal after\
            being in a state usable for data streaming.

        :param ~spinnman.transceiver.Transceiver transceiver:
            the SpiNNMan instance
        :param list(ExtraMonitorSupportMachineVertex) extra_monitor_cores:
            the extra monitor cores to set
        :param ~pacman.model.placements.Placements placements:
            placements object
        """
        # Set the routers to temporary values
        self.set_router_wait1_timeout(
            self._TEMP_TIMEOUT, transceiver, placements)
        self.set_router_wait2_timeout(
            self._ZERO_TIMEOUT, transceiver, placements)

        if self._last_status is None:
            log.warning(
                "Cores have not been set for data extraction, so can't be"
                " unset")
        try:
            self.set_router_wait1_timeout(
                self._last_status.router_wait1_timeout_parameters,
                transceiver, placements)
            self.set_router_wait2_timeout(
                self._last_status.router_wait2_timeout_parameters,
                transceiver, placements)

            lead_monitor = extra_monitor_cores[0]
            lead_monitor.set_reinjection_packets(
                placements, extra_monitor_cores, transceiver,
                point_to_point=self._last_status.is_reinjecting_point_to_point,
                multicast=self._last_status.is_reinjecting_multicast,
                nearest_neighbour=(
                    self._last_status.is_reinjecting_nearest_neighbour),
                fixed_route=self._last_status.is_reinjecting_fixed_route)
        except Exception:  # pylint: disable=broad-except
            log.exception("Error resetting timeouts")
            log.error("Checking if the cores are OK...")
            core_subsets = convert_vertices_to_core_subset(
                extra_monitor_cores, placements)
            try:
                error_cores = transceiver.get_cores_not_in_state(
                    core_subsets, {CPUState.RUNNING})
                if error_cores:
                    log.error("Cores in an unexpected state: {}".format(
                        error_cores))
            except Exception:  # pylint: disable=broad-except
                log.exception("Couldn't get core state")

    def __reprogram_tag(self, connection):
        """ Make our tag deliver to the given connection.

        :param ~.SCAMPConnection connection: The connection to deliver to.
        """
        request = IPTagSet(
            self._x, self._y, [0, 0, 0, 0], 0,
            self._remote_tag, strip=True, use_sender=True)
        data = connection.get_scp_data(request)
        exn = None
        for _ in range(3):
            try:
                connection.send(data)
                _, _, response, offset = connection.receive_scp_response()
                request.get_scp_response().read_bytestring(response, offset)
                return
            except SpinnmanTimeoutException as e:
                exn = e
        raise exn or SpinnmanTimeoutException

    def get_data(
            self, extra_monitor, extra_monitor_placement, memory_address,
            length_in_bytes, fixed_routes):
        """ Gets data from a given core and memory address.

        :param ExtraMonitorSupportMachineVertex extra_monitor:
            the extra monitor used for this data
        :param ~pacman.model.placements.Placement extra_monitor_placement:
            placement object for where to get data from
        :param int memory_address: the address in SDRAM to start reading from
        :param int length_in_bytes: the length of data to read in bytes
        :param fixed_routes: the fixed routes, used in the report of which
            chips were used by the speed up process
        :type fixed_routes: dict(tuple(int,int),~spinn_machine.FixedRouteEntry)
        :return: byte array of the data
        :rtype: bytearray
        """
        start = float(time.time())

        # if asked for no data, just return a empty byte array
        if length_in_bytes == 0:
            data = bytearray(0)
            end = float(time.time())
            self._provenance_data_items[
                extra_monitor_placement, memory_address,
                length_in_bytes].append((end - start, [0]))
            return data

        transceiver = get_simulator().transceiver

        # Update the IP Tag to work through a NAT firewall
        connection = SCAMPConnection(
            chip_x=self._x, chip_y=self._y, remote_host=self._ip_address)
        self.__reprogram_tag(connection)

        # update transaction id for extra monitor
        extra_monitor.update_transaction_id()
        transaction_id = extra_monitor.transaction_id

        # send
        connection.send_sdp_message(self.__make_sdp_message(
            extra_monitor_placement,
            SDP_PORTS.EXTRA_MONITOR_CORE_DATA_SPEED_UP,
            _FOUR_WORDS.pack(
                DATA_OUT_COMMANDS.START_SENDING.value, transaction_id,
                memory_address, length_in_bytes)))

        # receive
        self._output = bytearray(length_in_bytes)
        self._view = memoryview(self._output)
        self._max_seq_num = self.calculate_max_seq_num()
        lost_seq_nums = self._receive_data(
            transceiver, extra_monitor_placement, connection, transaction_id)

        # Stop anything else getting through (and reduce traffic)
        connection.send_sdp_message(self.__make_sdp_message(
            extra_monitor_placement,
            SDP_PORTS.EXTRA_MONITOR_CORE_DATA_SPEED_UP,
            _TWO_WORDS.pack(DATA_OUT_COMMANDS.CLEAR.value, transaction_id)))
        connection.close()

        end = float(time.time())
        self._provenance_data_items[
            extra_monitor_placement, memory_address, length_in_bytes].append(
                (end - start, lost_seq_nums))

        # create report elements
        if self._write_data_speed_up_reports:
            routers_been_in_use = self._determine_which_routers_were_used(
                extra_monitor_placement, fixed_routes,
                transceiver.get_machine_details())
            self._write_routers_used_into_report(
                self._out_report_path, routers_been_in_use,
                extra_monitor_placement)

        return self._output

    def _receive_data(
            self, transceiver, placement, connection, transaction_id):
        """
        :param ~.Transceiver transceiver:
        :param ~.Placement placement:
        :param ~.UDPConnection connection:
        :param int transaction_id:
        :rtype: list(int)
        """
        seq_nums = set()
        lost_seq_nums = list()
        timeoutcount = 0
        finished = False
        while not finished:
            try:
                data = connection.receive(
                    timeout=self._TIMEOUT_PER_RECEIVE_IN_SECONDS)
                response_transaction_id, = _ONE_WORD.unpack_from(data, 4)
                if transaction_id == response_transaction_id:
                    timeoutcount = 0
                    seq_nums, finished = self._process_data(
                        data, seq_nums, finished, placement, transceiver,
                        lost_seq_nums, transaction_id)
                else:
                    log.info(
                        "ignoring packet as transaction id should be {}"
                        " but is {}".format(
                            transaction_id, response_transaction_id))
            except SpinnmanTimeoutException as e:
                if timeoutcount > TIMEOUT_RETRY_LIMIT:
                    raise SpinnFrontEndException(
                        "Failed to hear from the machine during {} attempts. "
                        "Please try removing firewalls".format(
                            timeoutcount)) from e

                timeoutcount += 1
                # self.__reset_connection()
                if not finished:
                    finished = self._determine_and_retransmit_missing_seq_nums(
                        seq_nums, transceiver, placement, lost_seq_nums,
                        transaction_id)
        return lost_seq_nums

    @staticmethod
    def _determine_which_routers_were_used(placement, fixed_routes, machine):
        """ Traverse the fixed route paths from a given location to its\
            destination. Used for determining which routers were used.

        :param ~.Placement placement: the source to start from
        :param dict(tuple(int,int),~.MulticastRoutingEntry) fixed_routes:
            the fixed routes for each router
        :param ~.Machine machine: the spinnMachine instance
        :return: list of chip locations
        :rtype: list(tuple(int,int))
        """
        routers = [(placement.x, placement.y)]
        entry = fixed_routes[placement.x, placement.y]
        chip_x = placement.x
        chip_y = placement.y
        while len(entry.processor_ids) == 0:
            # can assume one link, as its a minimum spanning tree going to
            # the root
            machine_link = machine.get_chip_at(
                chip_x, chip_y).router.get_link(next(iter(entry.link_ids)))
            chip_x = machine_link.destination_x
            chip_y = machine_link.destination_y
            routers.append((chip_x, chip_y))
            entry = fixed_routes[chip_x, chip_y]
        return routers

    @staticmethod
    def _write_routers_used_into_report(
            report_path, routers_been_in_use, placement):
        """ Write the used routers into a report

        :param str report_path: the path to the report file
        :param list(tuple(int,int)) routers_been_in_use:
            the routers been in use
        :param ~.Placement placement: the first placement used
        """
        writer_behaviour = "w"
        if os.path.isfile(report_path):
            writer_behaviour = "a"

        with open(report_path, writer_behaviour) as writer:
            writer.write("[{}:{}:{}] = {}\n".format(
                placement.x, placement.y, placement.p, routers_been_in_use))

    def _calculate_missing_seq_nums(self, seq_nums):
        """ Determine which sequence numbers we've missed

        :param set(int) seq_nums: the set already acquired
        :return: list of missing sequence numbers
        :rtype: list(int)
        """
        return [sn for sn in range(self._max_seq_num) if sn not in seq_nums]

    def _determine_and_retransmit_missing_seq_nums(
            self, seq_nums, transceiver, placement, lost_seq_nums,
            transaction_id):
        """ Determine if there are any missing sequence numbers, and if so\
            retransmits the missing sequence numbers back to the core for\
            retransmission.

        :param set(int) seq_nums: the sequence numbers already received
        :param ~.Transceiver transceiver: spinnman instance
        :param ~.Placement placement: placement instance
        :param list(int) lost_seq_nums:
        :param int transaction_id: transaction_id
        :return: whether all packets are transmitted
        :rtype: bool
        """
        # pylint: disable=too-many-locals

        # locate missing sequence numbers from pile
        missing_seq_nums = self._calculate_missing_seq_nums(seq_nums)

        lost_seq_nums.append(len(missing_seq_nums))
        # self._print_missing(missing_seq_nums)
        if not missing_seq_nums:
            return True

        # figure n packets given the 2 formats
        n_packets = 1
        length_via_format2 = len(missing_seq_nums) - (
            WORDS_PER_FULL_PACKET - WORDS_FOR_COMMAND_N_MISSING_TRANSACTION)
        if length_via_format2 > 0:
            n_packets += ceildiv(
                length_via_format2,
                WORDS_PER_FULL_PACKET - WORDS_FOR_COMMAND_TRANSACTION)
        # self._print_missing_n_packets(n_packets)

        # transmit missing sequence as a new SDP packet
        first = True
        seq_num_offset = 0
        for _ in range(n_packets):
            length_left_in_packet = WORDS_PER_FULL_PACKET
            offset = 0

            # if first, add n packets to list
            if first:

                # get left over space / data size
                size_of_data_left_to_transmit = min(
                    length_left_in_packet -
                    WORDS_FOR_COMMAND_N_MISSING_TRANSACTION,
                    len(missing_seq_nums) - seq_num_offset)

                # build data holder accordingly
                data = bytearray(
                    (size_of_data_left_to_transmit +
                     WORDS_FOR_COMMAND_N_MISSING_TRANSACTION) * BYTES_PER_WORD)

                # pack flag and n packets
                _ONE_WORD.pack_into(
                    data, offset, DATA_OUT_COMMANDS.START_MISSING_SEQ.value)
                _TWO_WORDS.pack_into(
                    data, BYTES_PER_WORD, transaction_id, n_packets)

                # update state
                offset += (
                    WORDS_FOR_COMMAND_N_MISSING_TRANSACTION * BYTES_PER_WORD)
                length_left_in_packet -= (
                    WORDS_FOR_COMMAND_N_MISSING_TRANSACTION)
                first = False

            else:  # just add data
                # get left over space / data size
                size_of_data_left_to_transmit = min(
                    WORDS_PER_FULL_PACKET_WITH_SEQUENCE_NUM,
                    len(missing_seq_nums) - seq_num_offset)

                # build data holder accordingly
                data = bytearray(
                    (size_of_data_left_to_transmit +
                     WORDS_FOR_COMMAND_TRANSACTION) * BYTES_PER_WORD)

                # pack flag
                _TWO_WORDS.pack_into(
                    data, offset,
                    DATA_OUT_COMMANDS.MISSING_SEQ.value, transaction_id)
                offset += BYTES_PER_WORD * WORDS_FOR_COMMAND_TRANSACTION
                length_left_in_packet -= WORDS_FOR_COMMAND_TRANSACTION

            # fill data field
            n_word_struct(size_of_data_left_to_transmit).pack_into(
                data, offset, *missing_seq_nums[
                    seq_num_offset:
                    seq_num_offset + size_of_data_left_to_transmit])
            seq_num_offset += length_left_in_packet

            # build SDP message and send it to the core
            transceiver.send_sdp_message(self.__make_sdp_message(
                placement, SDP_PORTS.EXTRA_MONITOR_CORE_DATA_SPEED_UP, data))

            # sleep for ensuring core doesn't lose packets
            time.sleep(self._TIMEOUT_FOR_SENDING_IN_SECONDS)
            # self._print_packet_num_being_sent(packet_count, n_packets)
        return False

    def _process_data(
            self, data, seq_nums, finished, placement, transceiver,
            lost_seq_nums, transaction_id):
        """ Take a packet and processes it see if we're finished yet.

        :param bytearray data: the packet data
        :param set(int) seq_nums: the list of sequence numbers received so far
        :param bool finished: bool which states if finished or not
        :param ~.Placement placement:
            placement object for location on machine
        :param ~.Transceiver transceiver: spinnman instance
        :param int transaction_id: the transaction ID for this stream
        :param list(int) lost_seq_nums:
            the list of n sequence numbers lost per iteration
        :return: set of data items, if its the first packet, the list of
            sequence numbers, the sequence number received and if its finished
        :rtype: tuple(set(int), bool)
        """
        # pylint: disable=too-many-arguments
        # self._print_out_packet_data(data)
        length_of_data = len(data)
        first_packet_element, = _ONE_WORD.unpack_from(data, 0)

        # get flags
        seq_num = first_packet_element & self._SEQUENCE_NUMBER_MASK
        is_end_of_stream = (
            first_packet_element & self._LAST_MESSAGE_FLAG_BIT_MASK) != 0

        # check seq num not insane
        if seq_num > self._max_seq_num:
            raise Exception(
                "got an insane sequence number. got {} when "
                "the max is {} with a length of {}".format(
                    seq_num, self._max_seq_num, length_of_data))

        # figure offset for where data is to be put
        offset = self._calculate_offset(seq_num)

        # write data

        # read offset from data is at byte 8. as first 4 is seq num,
        # second 4 is transaction id
        true_data_length = (
                offset + length_of_data - BYTES_FOR_SEQ_AND_TRANSACTION_ID)
        if (not is_end_of_stream or
                length_of_data != BYTES_FOR_SEQ_AND_TRANSACTION_ID):
            self._write_into_view(
                offset, true_data_length, data,
                BYTES_FOR_SEQ_AND_TRANSACTION_ID, length_of_data, seq_num,
                length_of_data, False)

        # add seq num to list
        seq_nums.add(seq_num)

        # if received a last flag on its own, its during retransmission.
        #  check and try again if required
        if is_end_of_stream:
            if not self._check(seq_nums):
                finished = self._determine_and_retransmit_missing_seq_nums(
                    placement=placement, transceiver=transceiver,
                    seq_nums=seq_nums, lost_seq_nums=lost_seq_nums,
                    transaction_id=transaction_id)
            else:
                finished = True
        return seq_nums, finished

    @staticmethod
    def _calculate_offset(seq_num):
        """
        :param int seq_num:
        :rtype: int
        """
        return (seq_num * WORDS_PER_FULL_PACKET_WITH_SEQUENCE_NUM *
                BYTES_PER_WORD)

    def _write_into_view(
            self, view_start_position, view_end_position,
            data, data_start_position, data_end_position, seq_num,
            packet_length, is_final):
        """ Puts data into the view

        :param int view_start_position: where in view to start
        :param int view_end_position: where in view to end
        :param bytearray data: the data holder to write from
        :param int data_start_position: where in data holder to start from
        :param int data_end_position: where in data holder to end
        :param int seq_num: the sequence number to figure
        :raises Exception: If the position to write to is crazy
        """
        # pylint: disable=too-many-arguments
        if view_end_position > len(self._output):
            raise Exception(
                "I'm trying to add to my output data, but am trying to add "
                "outside my acceptable output positions! max is {} and I "
                "received a request to fill to {} from {} for sequence num "
                "{} from max sequence num {} length of packet {} and "
                "final {}".format(
                    len(self._output), view_end_position, view_start_position,
                    seq_num, self._max_seq_num - 1, packet_length, is_final))
        self._view[view_start_position: view_end_position] = \
            data[data_start_position:data_end_position]

    def _check(self, seq_nums):
        """ Verify if the sequence numbers are correct.

        :param list(int) seq_nums: the received sequence numbers
        :return: Whether all the sequence numbers have been received
        :rtype: bool
        """
        # hand back
        seq_nums = sorted(seq_nums)
        max_needed = self.calculate_max_seq_num()
        if len(seq_nums) > max_needed + 1:
            raise Exception("I've received more data than I was expecting!!")
        return len(seq_nums) == max_needed + 1

    def calculate_max_seq_num(self):
        """ Deduce the max sequence number expected to be received

        :return: the biggest sequence num expected
        :rtype: int
        """
        return ceildiv(
            len(self._output),
            WORDS_PER_FULL_PACKET_WITH_SEQUENCE_NUM * BYTES_PER_WORD)

    @staticmethod
    def _print_missing(seq_nums):
        """ Debug printer for the missing sequence numbers from the pile.

        :param list(int) seq_nums: the sequence numbers received so far
        """
        for seq_num in sorted(seq_nums):
            log.info("from list I'm missing sequence num {}", seq_num)

    @staticmethod
    def _print_missing_n_packets(n_packets):
        """ Debug printer for the number of missing packets from the pile.

        :param n_packets: the number of packets
        """
        log.info("missing packets = {}", n_packets)

    @staticmethod
    def _print_out_packet_data(data, position):
        """ Debug prints out the data from the packet.

        :param bytearray data: the packet data
        """
        reread_data = struct.unpack("<{}B".format(len(data)), data)
        output = ""
        position2 = position
        log.debug("size of data is {}".format((len(data) / 4) - 3))
        for index, reread_data_element in enumerate(reread_data):
            if index >= 12:
                output += "{}:{},".format(position2, reread_data_element)
                position2 += 1
        log.debug("converted data back into readable form is {}", output)

    @staticmethod
    def _print_length_of_received_seq_nums(seq_nums, max_needed):
        """ Debug helper method for figuring out if everything been received.

        :param list(int) seq_nums: sequence numbers received
        :param int max_needed: biggest expected to have
        """
        if len(seq_nums) != max_needed:
            log.info("should have received {} sequence numbers, but received "
                     "{} sequence numbers", max_needed, len(seq_nums))

    @staticmethod
    def _print_packet_num_being_sent(packet_count, n_packets):
        """ Debug helper for printing missing sequence number packet\
            transmission.

        :param int packet_count: which packet is being fired
        :param int n_packets: how many packets to fire.
        """
        log.debug("send SDP packet with missing sequence numbers: {} of {}",
                  packet_count + 1, n_packets)


class _StreamingContextManager(object):
    """ The implementation of the context manager object for streaming \
        configuration control.
    """
    __slots__ = ["_gatherers", "_monitors", "_placements", "_txrx"]

    def __init__(self, gatherers, txrx, monitors, placements):
        """
        :param iterable(DataSpeedUpPacketGatherMachineVertex) gatherers:
        :param ~spinnman.transceiver.Transceiver txrx:
        :param list(ExtraMonitorSupportMachineVertex) monitors:
        :param ~pacman.model.placements.Placements placements:
        """
        self._gatherers = list(gatherers)
        self._txrx = txrx
        self._monitors = monitors
        self._placements = placements

    def __enter__(self):
        for gatherer in self._gatherers:
            gatherer.load_system_routing_tables(
                self._txrx, self._monitors, self._placements)
        for gatherer in self._gatherers:
            gatherer.set_cores_for_data_streaming(
                self._txrx, self._monitors, self._placements)

    def __exit__(self, _type, _value, _tb):
        for gatherer in self._gatherers:
            gatherer.unset_cores_for_data_streaming(
                self._txrx, self._monitors, self._placements)
        for gatherer in self._gatherers:
            gatherer.load_application_routing_tables(
                self._txrx, self._monitors, self._placements)
        return False



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from spinn_front_end_common.interface.provenance import SqlLiteDatabase
from spinn_front_end_common.utilities.constants import PROVENANCE_DB


class ProvenanceSQLWriter(object):
    """ Writes provenance in SQL format.
    """

    __slots__ = []

    def __call__(self, provenance_data_items, provenance_data_path):
        """ Writes provenance in SQL format

        :param list(ProvenanceDataItem) provenance_data_items:
        :param str provenance_data_path:
        """
        database_file = os.path.join(
            provenance_data_path, PROVENANCE_DB)
        with SqlLiteDatabase(database_file) as db:
            db.insert_items(provenance_data_items)



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import struct
from spinn_utilities.log import FormatAdapter
from spinn_utilities.progress_bar import ProgressBar
from spinn_utilities.executable_finder import ExecutableFinder
from spinn_machine import CoreSubsets, Router
from spinnman.model import ExecutableTargets
from spinnman.model.enums import CPUState
from spinn_front_end_common.utilities.exceptions import SpinnFrontEndException
from spinn_front_end_common.utilities.system_control_logic import (
    run_system_application)
from spinn_front_end_common.utilities.utility_objs import ExecutableType
from spinn_front_end_common.utilities.helpful_functions import (
    get_defaultable_source_id)
_FOUR_WORDS = struct.Struct("<IIII")
_THREE_WORDS = struct.Struct("<III")
# The SDRAM Tag used by the application - note this is fixed in the APLX
_SDRAM_TAG = 1

logger = FormatAdapter(logging.getLogger(__name__))


def mundy_on_chip_router_compression(
        routing_tables, transceiver, machine, app_id,
        system_provenance_folder, write_compressor_iobuf,
        compress_only_when_needed=True, compress_as_much_as_possible=False):
    """ Load routing tables and compress them using Andrew Mundy's algorithm.

    This uses an APLX built by Mundy which no longer compiles but still works
    with the current tool chain

    :param ~pacman.model.routing_tables.MulticastRoutingTables routing_tables:
        the memory routing tables to be compressed
    :param ~spinnman.transceiver.Transceiver transceiver:
        How to talk to the machine
    :param ~spinn_machine.Machine machine:
        the SpiNNaker machine representation
    :param int app_id: the application ID used by the main application
    :param str system_provenance_folder: the path to where to write the data
    :param bool compress_as_much_as_possible:
        If False, the compressor will only reduce the table until it fits in
        the router space, otherwise it will try to reduce until it until it
        can't reduce it any more
    :param bool compress_only_when_needed:
        If True, the compressor will only compress if the table doesn't fit in
        the current router space, otherwise it will just load the table
    :param bool write_compressor_iobuf: Should IOBUF be read and written out
    :raises SpinnFrontEndException: If compression fails
    """
    # pylint: disable=too-many-arguments, unused-argument
    msg = (
        "MundyOnChipRouterCompression is no longer supported. "
        "To use the currently recommended compression algorithm remove "
        "loading_algorithms from your cfg. "
        "While not recommended, OrderedCoveringOnChipRouterCompression "
        "provides the same algorithm but has been updated to use the "
        "current tools.")
    print(msg)
    logger.warning(msg)
    raise NotImplementedError(msg)


def pair_compression(
        routing_tables, transceiver, executable_finder,
        machine, app_id, provenance_file_path, write_compressor_iobuf,
        compress_as_much_as_possible=True):
    """ Load routing tables and compress then using the Pair Algorithm.

    See ``pacman/operations/router_compressors/pair_compressor.py`` which is
    the exact same algorithm implemented in Python.

    :param ~pacman.model.routing_tables.MulticastRoutingTables routing_tables:
        the memory routing tables to be compressed
    :param ~spinnman.transceiver.Transceiver transceiver:
        How to talk to the machine
    :param executable_finder: tracker of binaries.
    :type executable_finder:
        ~spinn_utilities.executable_finder.ExecutableFinder
    :param ~spinn_machine.Machine machine:
        the SpiNNaker machine representation
    :param int app_id: the application ID used by the main application
    :param str provenance_file_path: the path to where to write the data
    :param bool compress_as_much_as_possible:
        If False, the compressor will only reduce the table until it fits in
        the router space, otherwise it will try to reduce until it until it
        can't reduce it any more
    :param bool write_compressor_iobuf: Should IOBUF be read and written out
    :raises SpinnFrontEndException: If compression fails
     """
    # pylint: disable=too-many-arguments
    binary_path = executable_finder.get_executable_path(
        "simple_pair_compressor.aplx")
    compression = Compression(
        app_id, binary_path, compress_as_much_as_possible,
        machine, provenance_file_path, routing_tables, transceiver,
        "Running pair routing table compression on chip",
        write_compressor_iobuf, result_register=1)
    compression.compress()


def ordered_covering_compression(
        routing_tables, transceiver, executable_finder,
        machine, app_id, provenance_file_path, write_compressor_iobuf,
        compress_as_much_as_possible=True):
    """ Load routing tables and compress then using the unordered Algorithm.

    To the best of our knowledge this is the same algorithm as
    :py:func:`mundy_on_chip_router_compression`, except this one is still
    buildable and can be maintained.

    :param ~pacman.model.routing_tables.MulticastRoutingTables routing_tables:
        the memory routing tables to be compressed
    :param ~spinnman.transceiver.Transceiver transceiver:
        How to talk to the machine
    :param executable_finder: tracker of binaries.
    :type executable_finder:
        ~spinn_utilities.executable_finder.ExecutableFinder
    :param ~spinn_machine.Machine machine:
        the SpiNNaker machine representation
    :param int app_id: the application ID used by the main application
    :param str provenance_file_path: the path to where to write the data
    :param bool compress_as_much_as_possible:
        If False, the compressor will only reduce the table until it fits in
        the router space, otherwise it will try to reduce until it until it
        can't reduce it any more
    :param bool write_compressor_iobuf: Should IOBUF be read and written out
    :raises SpinnFrontEndException: If compression fails
    """
    # pylint: disable=too-many-arguments
    binary_path = executable_finder.get_executable_path(
        "simple_unordered_compressor.aplx")
    compression = Compression(
        app_id, binary_path, compress_as_much_as_possible,
        machine, provenance_file_path, routing_tables, transceiver,
        "Running unordered routing table compression on chip",
        write_compressor_iobuf, result_register=1)
    compression.compress()


def unordered_compression(
        routing_tables, transceiver, executable_finder,
        machine, app_id, provenance_file_path, write_compressor_iobuf,
        compress_as_much_as_possible=True):
    """ DEPRECATED use ordered_covering_compression """
    logger.warning(
        "UnorderedOnChipRouterCompression algorithm name is deprecated. "
        "Please use OrderedCoveringOnChipRouterCompression instead. "
        "loading_algorithms from your cfg to use defaults")
    ordered_covering_compression(
        routing_tables, transceiver, executable_finder,
        machine, app_id, provenance_file_path, write_compressor_iobuf,
        compress_as_much_as_possible)


class Compression(object):
    """ Compression algorithm implementation that uses a on-chip router\
        compressor in order to parallelise.
    """

    __slots__ = [
        "_app_id",
        "_binary_path",
        "_compress_as_much_as_possible",
        "_compress_only_when_needed",
        "_compressor_app_id",
        "_machine",
        "_progresses_text",
        "_provenance_file_path",
        "__result_register",
        "_transceiver",
        "_routing_tables",
        "_write_compressor_iobuf",
        "__failures"]

    def __init__(
            self, app_id, binary_path, compress_as_much_as_possible,
            machine, provenance_file_path, routing_tables, transceiver,
            progress_text, write_compressor_iobuf, result_register):
        """
        :param int app_id: the application ID used by the main application
        :param str binary_path: What binary to run
        :param bool compress_as_much_as_possible:
            Whether to do maximal compression
        :param ~spinn_machine.Machine machine: The machine model
        :param str provenance_file_path:
            Where to write provenance data (IOBUF contents)
        :param routing_tables: the memory routing tables to be compressed
        :type routing_tables:
            ~pacman.model.routing_tables.MulticastRoutingTables
        :param ~spinnman.transceiver.Transceiver transceiver:
            How to talk to the machine
        :param str progress_text: Text to use in progress bar
        :param bool write_compressor_iobuf:
            Should IOBUF be read and written out
        :param int result_register:
            number of the user register to check for the result code
        """
        self._app_id = app_id
        self._binary_path = binary_path
        self._compress_as_much_as_possible = compress_as_much_as_possible
        # Only used by mundy compressor we can not rebuild
        self._compress_only_when_needed = None
        self._machine = machine
        self._provenance_file_path = provenance_file_path
        self._transceiver = transceiver
        self._routing_tables = routing_tables
        self._progresses_text = progress_text
        self._compressor_app_id = None
        self._write_compressor_iobuf = write_compressor_iobuf
        self.__failures = []
        self.__result_register = result_register

    def compress(self):
        """ Apply the on-machine compression algorithm.

        :raises SpinnFrontEndException: If compression fails
        """
        # pylint: disable=too-many-arguments

        # build progress bar
        progress_bar = ProgressBar(
            len(self._routing_tables.routing_tables) * 2,
            self._progresses_text)

        if len(self._routing_tables.routing_tables) == 0:
            progress_bar.end()
            return

        self._compressor_app_id = self._transceiver.app_id_tracker.get_new_id()

        # figure size of SDRAM needed for each chip for storing the routing
        # table
        for routing_table in progress_bar.over(self._routing_tables, False):
            self._load_routing_table(routing_table)

        # load the router compressor executable
        executable_targets = self._load_executables()

        executable_finder = ExecutableFinder(binary_search_paths=[])
        run_system_application(
            executable_targets, self._compressor_app_id, self._transceiver,
            self._provenance_file_path, executable_finder,
            self._write_compressor_iobuf,
            self._check_for_success,
            [CPUState.FINISHED], False, "compressor_on_{}_{}_{}.txt",
            [self._binary_path], progress_bar)
        if self.__failures:
            raise SpinnFrontEndException(
                "The router compressor failed on {}".format(self.__failures))

    def _load_routing_table(self, table):
        """
        :param pacman.model.routing_tables.AbstractMulticastRoutingTable table:
            the pacman router table instance
        """
        data = self._build_data(table)

        # go to spinnman and ask for a memory region of that size per chip.
        base_address = self._transceiver.malloc_sdram(
            table.x, table.y, len(data), self._compressor_app_id, _SDRAM_TAG)

        # write SDRAM requirements per chip
        self._transceiver.write_memory(table.x, table.y, base_address, data)

    def _check_for_success(self, executable_targets, transceiver):
        """ Goes through the cores checking for cores that have failed to\
            compress the routing tables to the level where they fit into the\
            router

        :param ExecutableTargets executable_targets:
        :param ~spinnman.transceiver.Transceiver transceiver:
        """
        for core_subset in executable_targets.all_core_subsets:
            x = core_subset.x
            y = core_subset.y
            for p in core_subset.processor_ids:
                # Read the result from specified register
                if self.__result_register == 0:
                    result = transceiver.read_user_0(x, y, p)
                elif self.__result_register == 1:
                    result = transceiver.read_user_1(x, y, p)
                elif self.__result_register == 2:
                    result = transceiver.read_user_2(x, y, p)
                else:
                    raise Exception("Incorrect register")
                # The result is 0 if success, otherwise failure
                if result != 0:
                    self.__failures.append((x, y))
        return len(self.__failures) == 0

    def _load_executables(self):
        """ Plans the loading of the router compressor onto the chips.

        :return:
            the executable targets that represent all cores/chips which have
            active routing tables
        :rtype: ExecutableTargets
        """

        # build core subsets
        core_subsets = CoreSubsets()
        for routing_table in self._routing_tables:

            # get the first none monitor core
            chip = self._machine.get_chip_at(routing_table.x, routing_table.y)
            processor = chip.get_first_none_monitor_processor()

            # add to the core subsets
            core_subsets.add_processor(
                routing_table.x, routing_table.y, processor.processor_id)

        # build executable targets
        executable_targets = ExecutableTargets()
        executable_targets.add_subsets(self._binary_path, core_subsets,
                                       ExecutableType.SYSTEM)

        return executable_targets

    def _build_data(self, table):
        """ Convert the router table into the data needed by the router\
            compressor C code.

        :param pacman.model.routing_tables.AbstractMulticastRoutingTable table:
            the pacman router table instance
        :return: The byte array of data
        :rtype: bytearray
        """

        # write header data of the app ID to load the data, if to store
        # results in SDRAM and the router table entries

        data = b''
        if self._compress_only_when_needed is None:
            data += _THREE_WORDS.pack(
                self._app_id,
                int(self._compress_as_much_as_possible),
                # Write the size of the table
                table.number_of_entries)
        else:
            # Mundy's compressor can not be changed so uses it own structure
            data += _FOUR_WORDS.pack(
                self._app_id, int(self._compress_only_when_needed),
                int(self._compress_as_much_as_possible),
                # Write the size of the table
                table.number_of_entries)

        for entry in table.multicast_routing_entries:
            data += _FOUR_WORDS.pack(
                entry.routing_entry_key, entry.mask,
                Router.convert_routing_table_entry_to_spinnaker_route(entry),
                get_defaultable_source_id(entry))
        return bytearray(data)



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinn_utilities.progress_bar import ProgressBar
from spinnman.messages.sdp import SDPHeader, SDPFlag
from spinnman.messages.scp.abstract_messages import AbstractSCPRequest
from spinnman.messages.scp import SCPRequestHeader
from spinnman.messages.scp.impl import CheckOKResponse
from spinnman.processes import AbstractMultiConnectionProcess
from spinn_front_end_common.utilities.constants import (
    SDP_PORTS, SDP_RUNNING_MESSAGE_CODES)


class _ClearIOBUFRequest(AbstractSCPRequest):
    def __init__(self, x, y, p):
        # pylint: disable=too-many-arguments

        super().__init__(
            SDPHeader(
                flags=SDPFlag.REPLY_EXPECTED,
                destination_cpu=p, destination_chip_x=x, destination_chip_y=y,
                destination_port=SDP_PORTS.RUNNING_COMMAND_SDP_PORT.value),
            SCPRequestHeader(
                command=SDP_RUNNING_MESSAGE_CODES.SDP_CLEAR_IOBUF_CODE),
            argument_3=int(True))

    def get_scp_response(self):
        return CheckOKResponse(
            "clear iobuf",
            SDP_RUNNING_MESSAGE_CODES.SDP_CLEAR_IOBUF_CODE.value)


class ClearIOBUFProcess(AbstractMultiConnectionProcess):
    """ How to clear the IOBUF buffers of a set of cores.

    .. note::
        The cores must be using the simulation interface.
    """

    def __init__(self, connection_selector):
        """
        :param connection_selector:
        :type connection_selector:
            ~spinnman.processes.abstract_multi_connection_process_connection_selector.AbstractMultiConnectionProcessConnectionSelector
        """
        super().__init__(connection_selector)
        self._progress = None

    def __receive_response(self, _response):
        if self._progress is not None:
            self._progress.update()

    def clear_iobuf(self, core_subsets, n_cores=None):
        """
        :param ~spinn_machine.CoreSubsets core_subsets:
        :param int n_cores: Defaults to the number of cores in `core_subsets`.
        """
        if n_cores is None:
            n_cores = len(core_subsets)
        self._progress = ProgressBar(
            n_cores, "clearing IOBUF from the machine")
        for core_subset in core_subsets:
            for processor_id in core_subset.processor_ids:
                self._send_request(
                    _ClearIOBUFRequest(
                        core_subset.x, core_subset.y, processor_id),
                    callback=self.__receive_response)
        self._finish()
        self._progress.end()
        self._progress = None
        self.check_for_error()



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinn_utilities.overrides import overrides
from spinnman.messages.scp import SCPRequestHeader
from spinnman.messages.scp.abstract_messages import (
    AbstractSCPRequest, AbstractSCPResponse)
from spinnman.messages.scp.enums import SCPResult
from spinnman.messages.sdp import SDPFlag, SDPHeader
from spinnman.exceptions import SpinnmanUnexpectedResponseCodeException
from spinn_front_end_common.utilities.constants import SDP_PORTS
from spinn_front_end_common.utilities.utility_objs import ReInjectionStatus
from .reinjector_scp_commands import ReinjectorSCPCommands


class GetReinjectionStatusMessage(AbstractSCPRequest):
    """ An SCP Request to get the status of the dropped packet reinjection.
    """

    __slots__ = []

    def __init__(self, x, y, p):
        """
        :param int x: The x-coordinate of a chip, between 0 and 255
        :param int y: The y-coordinate of a chip, between 0 and 255
        :param int p:
            The processor running the extra monitor vertex, between 0 and 17
        """
        super().__init__(
            SDPHeader(
                flags=SDPFlag.REPLY_EXPECTED,
                destination_port=(
                    SDP_PORTS.EXTRA_MONITOR_CORE_REINJECTION.value),
                destination_cpu=p, destination_chip_x=x,
                destination_chip_y=y),
            SCPRequestHeader(command=ReinjectorSCPCommands.GET_STATUS))

    @overrides(AbstractSCPRequest.get_scp_response)
    def get_scp_response(self):
        return GetReinjectionStatusMessageResponse(
            ReinjectorSCPCommands.GET_STATUS)


class GetReinjectionStatusMessageResponse(AbstractSCPResponse):
    """ An SCP response to a request for the dropped packet reinjection status
    """

    def __init__(self, command_code):
        super().__init__()
        self._reinjection_functionality_status = None
        self._command_code = command_code

    @overrides(AbstractSCPResponse.read_data_bytestring)
    def read_data_bytestring(self, data, offset):
        result = self.scp_response_header.result
        if result != SCPResult.RC_OK:
            raise SpinnmanUnexpectedResponseCodeException(
                "Get packet reinjection status", self._command_code,
                result.name)
        self._reinjection_functionality_status = \
            ReInjectionStatus(data, offset)

    @property
    def reinjection_functionality_status(self):
        return self._reinjection_functionality_status



# Copyright (c) 2019-2020 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from collections import defaultdict
import sys
from spinn_front_end_common.utilities.utility_objs import ExecutableType
from spinn_front_end_common.abstract_models import AbstractHasAssociatedBinary


class FindApplicationChipsUsed(object):
    """ Builds a set of stats on how many chips were used for application\
        cores.
    """

    def __call__(self, placements):
        """ Finds how many application chips there were and the cost on \
            each chip

        :param ~pacman.model.placements.Placements placements: placements
        :return: a tuple with 4 elements.

            1. how many chips were used
            2. the max application cores on any given chip
            3. the lowest number of application cores on any given chip
            4. the average number of application cores on any given chip

        :rtype: tuple(int,int,int,float)
        """
        chips_used = defaultdict(int)
        for placement in placements:
            # find binary type if applicable
            binary_start_type = None
            if isinstance(placement.vertex, AbstractHasAssociatedBinary):
                binary_start_type = placement.vertex.get_binary_start_type()
            if binary_start_type != ExecutableType.SYSTEM:
                chips_used[placement.x, placement.y] += 1

        low = sys.maxsize
        high = 0
        total = 0
        n_chips_used = len(chips_used)

        for key in chips_used:
            if chips_used[key] < low:
                low = chips_used[key]
            if chips_used[key] > high:
                high = chips_used[key]
            total += chips_used[key]
        average = total / n_chips_used
        return n_chips_used, high, low, average



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spalloc import ProtocolClient
from spinn_machine.virtual_machine import virtual_machine
from spinn_machine.machine import Machine


class SpallocMaxMachineGenerator(object):
    """ Generates a maximum virtual machine a given allocation server can\
        generate.
    """

    __slots__ = []

    def __call__(
            self, spalloc_server, spalloc_port=22244, spalloc_machine=None,
            max_sdram_size=None, max_machine_core_reduction=0):
        """
        :param str spalloc_server:
        :param int spalloc_port:
        :param str spalloc_machine:
        :param int max_sdram_size:
        :param int max_machine_core_reduction: the number of cores less than
            :py:const:`~spinn_machine.Machine.DEFAULT_MAX_CORES_PER_CHIP`
            that each chip should have
        :return: A virtual machine
        :rtype: ~spinn_machine.Machine
        """
        with ProtocolClient(spalloc_server, spalloc_port) as client:
            machines = client.list_machines()
            # Close the context immediately; don't want to keep this particular
            # connection around as there's not a great chance of this code
            # being rerun in this process any time soon.
        max_width = None
        max_height = None
        max_area = -1

        for machine in self._filter(machines, spalloc_machine):
            # Get the width and height in chips, and logical area in chips**2
            width, height, area = self._get_size(machine)

            # The "biggest" board is the one with the most chips
            if area > max_area:
                max_area = area
                max_width = width
                max_height = height

        if max_width is None:
            raise Exception(
                "The spalloc server appears to have no compatible machines")

        n_cpus_per_chip = (Machine.max_cores_per_chip() -
                           max_machine_core_reduction)

        # Return the width and height, and make no assumption about wrap-
        # arounds or version.
        return virtual_machine(
            width=max_width, height=max_height, sdram_per_chip=max_sdram_size,
            n_cpus_per_chip=n_cpus_per_chip, validate=False)

    @staticmethod
    def _filter(machines, target_name):
        """
        :param list(dict(str,str)) machines:
        :param str target_name:
        :rtype: iterable(dict(str,str or int))
        """
        if target_name is None:
            return (m for m in machines if "default" in m["tags"])
        return (m for m in machines if m["name"] == target_name)

    @staticmethod
    def _get_size(machine):
        """
        :param dict(str,int) machine:
        :return: width, height, area
        :rtype: tuple(int,int,int)
        """
        # Get the width and height in chips
        width = machine["width"] * 12
        height = machine["height"] * 12

        # A specific exception is the 1-board machine, which is represented as
        # a 3 board machine with 2 dead boards. In this case the width and
        # height are 8.
        if (machine["width"] == 1 and
                machine["height"] == 1 and
                len(machine["dead_boards"]) == 2):
            return 8, 8, 48

        return width, height, width * height



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


class DsWriteInfo(object):

    __slots__ = ["_db"]

    def __init__(self, database):
        """
        :param DsAbstractDatabase database: Database to map
        """
        # pylint: disable=super-init-not-called
        self._db = database

    def __getitem__(self, core):
        """
        Implements the mapping __getitem__ as long as core is the right type

        :param tuple(int,int,int) core:triple of (x, y, p)
        :return: dict with the keys
            ``start_address``, ``memory_used`` and ``memory_written``
        :rtype: dict(str,int)
        """
        (x, y, p) = core
        return self.get_info(x, y, p)

    def get_info(self, x, y, p):
        """
        gets the info for the core x, y, p

        :param int x: core x
        :param int y: core y
        :param int p: core p
        :return: dict with the keys
            ``start_address``, ``memory_used`` and ``memory_written``
        :rtype: dict(str,int)
        """
        return self._db.get_write_info(x, y, p)

    def set_info(self, x, y, p, info):
        """ Sets the info for the core x, y, p

        :param int x: core x
        :param int y: core y
        :param int p: core p
        :param dict(str,int) info: dict with the keys
            ``start_address``, ``memory_used`` and ``memory_written``
        """
        self._db.set_write_info(x, y, p, info)

    def set_size_info(self, x, y, p, memory_used):
        """ Sets the size info for the core x, y p.

        :param x: core x
        :param y: core y
        :param p: core p
        :param memory_used: memory allocated
        :rtype None:
        """
        self._db.set_size_info(x, y, p, memory_used)

    def clear_write_info(self):
        """ Clears the info for all cores,
        """
        self._db.clear_write_info()

    def keys(self):
        """ Yields the keys.

        As the more typical call is iteritems this makes use of that

        :rtype: iterable(tuple(int,int,int))
        """
        for key, _value in self._db.info_iteritems():
            yield key

    __iter__ = keys

    def __len__(self):
        """
        TEMP implementation

        :return:
        """
        return self._db.info_n_cores()

    def items(self):
        return self._db.info_iteritems()

    # Python 2 backward compatibility
    iteritems = items



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinn_utilities.progress_bar import ProgressBar
from spinn_front_end_common.interface.provenance import (
    AbstractProvidesLocalProvenanceData)


class GraphProvenanceGatherer(object):
    """ Gets provenance information from the graphs.
    """
    __slots__ = []

    def __call__(self, machine_graph, application_graph=None):
        """
        :param ~pacman.model.graphs.machine.MachineGraph machine_graph:
            The machine graph to inspect
        :param application_graph: The optional application graph
        :type application_graph:
            ~pacman.model.graphs.application.ApplicationGraph
        :rtype: list(ProvenanceDataItem)
        """
        prov_items = list()

        progress = ProgressBar(
            machine_graph.n_vertices +
            machine_graph.n_outgoing_edge_partitions,
            "Getting provenance data from machine graph")
        for vertex in progress.over(machine_graph.vertices, False):
            if isinstance(vertex, AbstractProvidesLocalProvenanceData):
                prov_items.extend(vertex.get_local_provenance_data())
        for partition in progress.over(machine_graph.outgoing_edge_partitions):
            for edge in partition.edges:
                if isinstance(edge, AbstractProvidesLocalProvenanceData):
                    prov_items.extend(edge.get_local_provenance_data())

        if application_graph is not None:
            progress = ProgressBar(
                application_graph.n_vertices +
                application_graph.n_outgoing_edge_partitions,
                "Getting provenance data from application graph")
            for vertex in progress.over(application_graph.vertices, False):
                if isinstance(vertex, AbstractProvidesLocalProvenanceData):
                    prov_items.extend(vertex.get_local_provenance_data())
            for partition in progress.over(
                    application_graph.outgoing_edge_partitions):
                for edge in partition.edges:
                    if isinstance(edge, AbstractProvidesLocalProvenanceData):
                        prov_items.extend(edge.get_local_provenance_data())

        return prov_items



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from spinn_utilities.require_subclass import require_subclass
from pacman.model.graphs.machine import MachineVertex


@require_subclass(MachineVertex)
class AbstractHasProfileData(object, metaclass=AbstractBase):
    """ Indicates a :py:class:`~pacman.model.graphs.machine.MachineVertex` \
        that can record a profile.
    """
    __slots__ = ()

    @abstractmethod
    def get_profile_data(self, transceiver, placement):
        """ Get the profile data recorded during simulation

        :param ~spinnman.transceiver.Transceiver transceiver:
        :param ~pacman.model.placements.Placement placement:
        :rtype: ~spinn_front_end_common.interface.profiling.ProfileData
        """



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from spinn_utilities.progress_bar import ProgressBar
from spinn_machine import SDRAM
from data_specification import DataSpecificationExecutor
from data_specification.constants import MAX_MEM_REGIONS
from data_specification.utility_calls import (
    get_region_base_address_offset, get_data_spec_and_file_writer_filename)
from spinn_front_end_common.abstract_models import (
    AbstractRewritesDataSpecification)
from spinn_front_end_common.utilities.helpful_functions import (
    generate_unique_folder_name, n_word_struct)


class DSGRegionReloader(object):
    """ Regenerates and reloads the data specifications.
    """
    __slots__ = [
        "_txrx", "_host", "_write_text", "_rpt_dir", "_data_dir"]

    def __call__(
            self, transceiver, placements, hostname, report_directory,
            write_text_specs):
        """
        :param ~spinnman.transceiver.Transceiver transceiver:
            SpiNNMan transceiver for communication
        :param ~pacman.model.placements.Placements placements:
            the list of placements of the machine graph to cores
        :param str hostname:
            the machine name
        :param str report_directory:
            the location where reports are stored
        :param bool write_text_specs:
            Whether the textual version of the specification is to be written
        """
        # pylint: disable=too-many-arguments, attribute-defined-outside-init
        self._txrx = transceiver
        self._host = hostname
        self._write_text = write_text_specs

        # build file paths for reloaded stuff
        app_data_dir = generate_unique_folder_name(
            report_directory, "reloaded_data_regions", "")
        if not os.path.exists(app_data_dir):
            os.makedirs(app_data_dir)
        self._data_dir = app_data_dir

        report_dir = None
        if write_text_specs:
            report_dir = generate_unique_folder_name(
                report_directory, "reloaded_data_regions", "")
            if not os.path.exists(report_dir):
                os.makedirs(report_dir)
        self._rpt_dir = report_dir

        progress = ProgressBar(placements.n_placements, "Reloading data")
        for placement in progress.over(placements.placements):
            # Generate the data spec for the placement if needed
            self._regenerate_data_spec_for_vertices(placement)

        # App data directory can be removed as should be empty
        os.rmdir(app_data_dir)

    def _regenerate_data_spec_for_vertices(self, placement):
        """
        :param ~.Placement placement:
        """
        vertex = placement.vertex

        # If the vertex doesn't regenerate, skip
        if not isinstance(vertex, AbstractRewritesDataSpecification):
            return

        # If the vertex doesn't require regeneration, skip
        if not vertex.reload_required():
            return

        # build the writers for the reports and data
        spec_file, spec = get_data_spec_and_file_writer_filename(
            placement.x, placement.y, placement.p, self._host,
            self._rpt_dir, self._write_text, self._data_dir)

        # Execute the regeneration
        vertex.regenerate_data_specification(spec, placement)

        # execute the spec
        with open(spec_file, "rb") as spec_reader:
            data_spec_executor = DataSpecificationExecutor(
                spec_reader, SDRAM.max_sdram_found)
            data_spec_executor.execute()
        try:
            os.remove(spec_file)
        except Exception:  # pylint: disable=broad-except
            # Ignore the deletion of files as non-critical
            pass

        # Read the region table for the placement
        regions_base_address = self._txrx.get_cpu_information_from_core(
            placement.x, placement.y, placement.p).user[0]
        start_region = get_region_base_address_offset(regions_base_address, 0)
        table_size = get_region_base_address_offset(
            regions_base_address, MAX_MEM_REGIONS) - start_region
        offsets = n_word_struct(MAX_MEM_REGIONS).unpack_from(
            self._txrx.read_memory(
                placement.x, placement.y, start_region, table_size))

        # Write the regions to the machine
        for i, region in enumerate(data_spec_executor.dsef.mem_regions):
            if region is not None and not region.unfilled:
                self._txrx.write_memory(
                    placement.x, placement.y, offsets[i],
                    region.region_data[:region.max_write_pointer])
        vertex.set_reload_required(False)



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from enum import Enum
from spinn_utilities.abstract_base import abstractproperty
from spinn_utilities.overrides import overrides
from data_specification.utility_calls import get_region_base_address_offset
from .abstract_provides_provenance_data_from_machine import (
    AbstractProvidesProvenanceDataFromMachine)
from spinn_front_end_common.utilities.constants import BYTES_PER_WORD
from spinn_front_end_common.utilities.helpful_functions import n_word_struct
from spinn_front_end_common.utilities.utility_objs import ProvenanceDataItem


def add_name(names, name):
    """
    :param iterable(str) names:
    :param str name:
    :rtype: list(str)
    """
    new_names = list(names)
    new_names.append(name)
    return new_names


class ProvidesProvenanceDataFromMachineImpl(
        AbstractProvidesProvenanceDataFromMachine, allow_derivation=True):
    """ An implementation that gets provenance data from a region of ints on\
        the machine.
    """

    __slots__ = ()

    class PROVENANCE_DATA_ENTRIES(Enum):
        """ Entries for the provenance data generated by models using provides\
            provenance vertex.
        """
        #: The counter of transmission overflows
        TRANSMISSION_EVENT_OVERFLOW = 0
        #: The counter of the number of times the callback queue was overloaded
        CALLBACK_QUEUE_OVERLOADED = 1
        #: The counter of the number of times the DMA queue was overloaded
        DMA_QUEUE_OVERLOADED = 2
        #: Whether the timer tick has overrun at all at any point
        TIMER_TIC_HAS_OVERRUN = 3
        #: The counter of the number of times the timer tick overran
        MAX_NUMBER_OF_TIMER_TIC_OVERRUN = 4

    NUM_PROVENANCE_DATA_ENTRIES = len(PROVENANCE_DATA_ENTRIES)

    _TIMER_TICK_OVERRUN = "Times_the_timer_tic_over_ran"
    _MAX_TIMER_TICK_OVERRUN = "max_number_of_times_timer_tic_over_ran"
    _TIMES_DMA_QUEUE_OVERLOADED = "Times_the_dma_queue_was_overloaded"
    _TIMES_TRANSMISSION_SPIKES_OVERRAN = \
        "Times_the_transmission_of_spikes_overran"
    _TIMES_CALLBACK_QUEUE_OVERLOADED = \
        "Times_the_callback_queue_was_overloaded"

    _TIMES_TRANSMISSION_SPIKES_OVERRAN_MESSAGE = (
        "The transmission buffer for {} on {}, {}, {} was blocked "
        "on {} occasions. This is often a sign that the system is "
        "experiencing back pressure from the communication fabric. "
        "Please either: "
        "1. spread the load over more cores, "
        "2. reduce your peak transmission load,"
        "3. adjust your mapping algorithm.")

    _TIMES_CALLBACK_QUEUE_OVERLOADED_MESSAGE = (
        "The callback queue for {} on {}, {}, {} overloaded on {} "
        "occasions. This is often a sign that the system is running "
        "too quickly for the number of neurons per core.  Please "
        "increase the machine time step or time_scale_factor or "
        "decrease the number of neurons per core.")

    _TIMES_DMA_QUEUE_OVERLOADED_MESSAGE = (
        "The DMA queue for {} on {}, {}, {} overloaded on {} "
        "occasions. This is often a sign that the system is running "
        "too quickly for the number of neurons per core.  Please "
        "increase the machine time step or time_scale_factor or "
        "decrease the number of neurons per core.")

    _TIMER_TICK_OVERRUN_MESSAGE = (
        "A Timer tick callback was still executing when the next "
        "timer tick callback was fired off for {} on {}, {}, {}, {} "
        "times. This is a sign of the system being overloaded and "
        "therefore the results are likely incorrect.  Please increase "
        "the machine time step or time_scale_factor or decrease the "
        "number of neurons per core")

    _MAX_TIMER_TICK_OVERRUN_MESSAGE = (
        "The timer for {} on {}, {}, {} fell behind by up to {} "
        "ticks. This is a sign of the system being overloaded and "
        "therefore the results are likely incorrect. Please increase "
        "the machine time step or time_scale_factor or decrease the "
        "number of neurons per core")

    @abstractproperty
    def _provenance_region_id(self):
        """
        :return: provenance_region_id
        :rtype: int
        """

    @abstractproperty
    def _n_additional_data_items(self):
        """
        :return: n_additional_data_items
        :rtype: int
        """

    def reserve_provenance_data_region(self, spec):
        """
        :param ~data_specification.DataSpecificationGenerator spec:
            The data specification being written.
        """
        spec.reserve_memory_region(
            self._provenance_region_id,
            self.get_provenance_data_size(self._n_additional_data_items),
            label="Provenance", empty=True)

    @classmethod
    def get_provenance_data_size(cls, n_additional_data_items):
        """
        :param int n_additional_data_items:
        :rtype: int
        """
        return (
            (cls.NUM_PROVENANCE_DATA_ENTRIES + n_additional_data_items)
            * BYTES_PER_WORD)

    def _get_provenance_region_address(self, transceiver, placement):
        """
        :param ~spinnman.transceiver.Transceiver transceiver:
        :param ~pacman.model.placements.Placement placement:
        :rtype: int
        """
        # Get the App Data for the core
        region_table_address = transceiver.get_cpu_information_from_core(
            placement.x, placement.y, placement.p).user[0]

        # Get the provenance region base address
        prov_region_entry_address = get_region_base_address_offset(
            region_table_address, self._provenance_region_id)
        return transceiver.read_word(
            placement.x, placement.y, prov_region_entry_address)

    def _read_provenance_data(self, transceiver, placement):
        """
        :param ~spinnman.transceiver.Transceiver transceiver:
        :param ~pacman.model.placements.Placement placement:
        :rtype: iterable(int)
        """
        provenance_address = self._get_provenance_region_address(
            transceiver, placement)
        data = transceiver.read_memory(
            placement.x, placement.y, provenance_address,
            self.get_provenance_data_size(self._n_additional_data_items))
        return n_word_struct(
            self.NUM_PROVENANCE_DATA_ENTRIES +
            self._n_additional_data_items).unpack_from(data)

    @staticmethod
    def _get_placement_details(placement):
        """
        :param ~pacman.model.placements.Placement placement:
        :rtype: tuple(str,int,int,list(str))
        """
        label = placement.vertex.label
        x = placement.x
        y = placement.y
        p = placement.p
        names = ["vertex_{}_{}_{}_{}".format(x, y, p, label)]
        return label, x, y, p, names

    @staticmethod
    def _add_name(names, name):
        """
        :param iterable(str) names:
        :param str name:
        :rtype: list(str)
        """
        return add_name(names, name)

    @staticmethod
    def _add_names(names, extra_names):
        """
        :param iterable(str) names:
        :param iterable(str) extra_names:
        :rtype: list(str)
        """
        new_names = list(names)
        new_names.extend(extra_names)
        return new_names

    def _read_basic_provenance_items(self, provenance_data, placement):
        """
        :param iterable(int) provenance_data:
        :param ~pacman.model.placements.Placement placement:
        :rtype: list(ProvenanceDataItem)
        """
        transmission_event_overflow = provenance_data[
            self.PROVENANCE_DATA_ENTRIES.TRANSMISSION_EVENT_OVERFLOW.value]
        callback_queue_overloaded = provenance_data[
            self.PROVENANCE_DATA_ENTRIES.CALLBACK_QUEUE_OVERLOADED.value]
        dma_queue_overloaded = provenance_data[
            self.PROVENANCE_DATA_ENTRIES.DMA_QUEUE_OVERLOADED.value]
        number_of_times_timer_tic_over_ran = provenance_data[
            self.PROVENANCE_DATA_ENTRIES.TIMER_TIC_HAS_OVERRUN.value]
        max_number_of_times_timer_tic_over_ran = provenance_data[
            self.PROVENANCE_DATA_ENTRIES.MAX_NUMBER_OF_TIMER_TIC_OVERRUN.value]

        # create provenance data items for returning
        label, x, y, p, names = self._get_placement_details(placement)
        data_items = list()
        data_items.append(ProvenanceDataItem(
            self._add_name(names, self._TIMES_TRANSMISSION_SPIKES_OVERRAN),
            transmission_event_overflow,
            report=transmission_event_overflow != 0,
            message=self._TIMES_TRANSMISSION_SPIKES_OVERRAN_MESSAGE.format(
                label, x, y, p, transmission_event_overflow)))

        data_items.append(ProvenanceDataItem(
            self._add_name(names, self._TIMES_CALLBACK_QUEUE_OVERLOADED),
            callback_queue_overloaded,
            report=callback_queue_overloaded != 0,
            message=self._TIMES_CALLBACK_QUEUE_OVERLOADED_MESSAGE.format(
                label, x, y, p, callback_queue_overloaded)))

        data_items.append(ProvenanceDataItem(
            self._add_name(names, self._TIMES_DMA_QUEUE_OVERLOADED),
            dma_queue_overloaded,
            report=dma_queue_overloaded != 0,
            message=self._TIMES_DMA_QUEUE_OVERLOADED_MESSAGE.format(
                label, x, y, p, dma_queue_overloaded)))

        data_items.append(ProvenanceDataItem(
            self._add_name(names, self._TIMER_TICK_OVERRUN),
            number_of_times_timer_tic_over_ran,
            report=number_of_times_timer_tic_over_ran != 0,
            message=self._TIMER_TICK_OVERRUN_MESSAGE.format(
                label, x, y, p, number_of_times_timer_tic_over_ran)))

        data_items.append(ProvenanceDataItem(
            self._add_name(names, self._MAX_TIMER_TICK_OVERRUN),
            max_number_of_times_timer_tic_over_ran,
            report=max_number_of_times_timer_tic_over_ran > 0,
            message=self._MAX_TIMER_TICK_OVERRUN_MESSAGE.format(
                label, x, y, p, max_number_of_times_timer_tic_over_ran)))

        return data_items

    def _get_remaining_provenance_data_items(self, provenance_data):
        """
        :param list(ProvenanceDataItem) provenance_data:
        :rtype: list(ProvenanceDataItem)
        """
        return provenance_data[self.NUM_PROVENANCE_DATA_ENTRIES:]

    @overrides(
        AbstractProvidesProvenanceDataFromMachine.
        get_provenance_data_from_machine,
        extend_doc=False)
    def get_provenance_data_from_machine(self, transceiver, placement):
        """ Retrieve the provenance data.

        :param ~spinnman.transceiver.Transceiver transceiver:
            How to talk to the machine
        :param ~pacman.model.placements.Placement placement:
            Which vertex are we retrieving from, and where was it
        :rtype:
            list(~spinn_front_end_common.utilities.utility_objs.ProvenanceDataItem)
        """
        provenance_data = self._read_provenance_data(
            transceiver, placement)
        return self._read_basic_provenance_items(provenance_data, placement)



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from spinn_utilities.require_subclass import require_subclass
from pacman.model.graphs.machine import MachineVertex


@require_subclass(MachineVertex)
class AbstractHasAssociatedBinary(object, metaclass=AbstractBase):
    """ Marks a machine graph vertex that can be launched on a SpiNNaker core.
    """

    __slots__ = ()

    @abstractmethod
    def get_binary_file_name(self):
        """ Get the binary name to be run for this vertex.

        :rtype: str
        """

    @abstractmethod
    def get_binary_start_type(self):
        """ Get the start type of the binary to be run.

        :rtype: ~spinn_front_end_common.utilities.utility_objs.ExecutableType
        """



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from enum import IntEnum
import struct
from spinn_utilities.overrides import overrides
from pacman.executor.injection_decorator import inject_items
from pacman.model.graphs.machine import MachineVertex
from pacman.model.resources import (
    ConstantSDRAM, CPUCyclesPerTickResource, DTCMResource, ResourceContainer)
from spinn_front_end_common.interface.provenance import (
    AbstractProvidesProvenanceDataFromMachine,
    ProvidesProvenanceDataFromMachineImpl)
from spinn_front_end_common.interface.simulation.simulation_utilities import (
    get_simulation_header_array)
from spinn_front_end_common.abstract_models import (
    AbstractGeneratesDataSpecification, AbstractHasAssociatedBinary,
    AbstractSupportsDatabaseInjection)
from spinn_front_end_common.utilities.utility_objs import (
    ProvenanceDataItem, ExecutableType)
from spinn_front_end_common.utilities.constants import (
    SYSTEM_BYTES_REQUIREMENT, SIMULATION_N_BYTES, BYTES_PER_WORD)
from spinn_front_end_common.utilities.exceptions import ConfigurationException

_ONE_SHORT = struct.Struct("<H")
_TWO_BYTES = struct.Struct("<BB")


class LivePacketGatherMachineVertex(
        MachineVertex, ProvidesProvenanceDataFromMachineImpl,
        AbstractGeneratesDataSpecification, AbstractHasAssociatedBinary,
        AbstractSupportsDatabaseInjection):
    """ Used to gather multicast packets coming from cores and stream them \
        out to a receiving application on host. Only ever deployed on chips \
        with a working Ethernet connection.
    """
    class _REGIONS(IntEnum):
        SYSTEM = 0
        CONFIG = 1
        PROVENANCE = 2

    #: Used to identify tags involved with the live packet gatherer.
    TRAFFIC_IDENTIFIER = "LPG_EVENT_STREAM"

    _N_ADDITIONAL_PROVENANCE_ITEMS = 2
    _CONFIG_SIZE = 12 * BYTES_PER_WORD
    _PROVENANCE_REGION_SIZE = 2 * BYTES_PER_WORD

    def __init__(
            self, lpg_params, constraints=None, app_vertex=None, label=None):
        """
        :param LivePacketGatherParameters lpg_params:
        :param LivePacketGather app_vertex:
        :param str label:
        :param constraints:
        :type constraints:
            iterable(~pacman.model.constraints.AbstractConstraint)
        """
        # inheritance
        super().__init__(
            label or lpg_params.label, constraints=constraints,
            app_vertex=app_vertex)

        self._resources_required = ResourceContainer(
            cpu_cycles=CPUCyclesPerTickResource(self.get_cpu_usage()),
            dtcm=DTCMResource(self.get_dtcm_usage()),
            sdram=ConstantSDRAM(self.get_sdram_usage()),
            iptags=[lpg_params.get_iptag_resource()])

        # app specific data items
        self._lpg_params = lpg_params

    @property
    @overrides(ProvidesProvenanceDataFromMachineImpl._provenance_region_id)
    def _provenance_region_id(self):
        return self._REGIONS.PROVENANCE

    @property
    @overrides(ProvidesProvenanceDataFromMachineImpl._n_additional_data_items)
    def _n_additional_data_items(self):
        return self._N_ADDITIONAL_PROVENANCE_ITEMS

    @property
    @overrides(MachineVertex.resources_required)
    def resources_required(self):
        return self._resources_required

    @property
    @overrides(AbstractSupportsDatabaseInjection.is_in_injection_mode)
    def is_in_injection_mode(self):
        return True

    @overrides(AbstractProvidesProvenanceDataFromMachine.
               get_provenance_data_from_machine)
    def get_provenance_data_from_machine(self, transceiver, placement):
        provenance_data = self._read_provenance_data(transceiver, placement)
        provenance_items = self._read_basic_provenance_items(
            provenance_data, placement)
        provenance_data = self._get_remaining_provenance_data_items(
            provenance_data)
        _, _, _, _, names = self._get_placement_details(placement)

        provenance_items.append(ProvenanceDataItem(
            self._add_name(names, "lost_packets_without_payload"),
            provenance_data[0],
            report=provenance_data[0] > 0,
            message=(
                "The live packet gatherer has lost {} packets which have "
                "payloads during its execution. Try increasing the machine "
                "time step or increasing the time scale factor. If you are "
                "running in real time, try reducing the number of vertices "
                "which are feeding this live packet gatherer".format(
                    provenance_data[0]))))
        provenance_items.append(ProvenanceDataItem(
            self._add_name(names, "lost_packets_with_payload"),
            provenance_data[1],
            report=provenance_data[1] > 0,
            message=(
                "The live packet gatherer has lost {} packets which do not "
                "have payloads during its execution. Try increasing the "
                "machine time step or increasing the time scale factor. If "
                "you are running in real time, try reducing the number of "
                "vertices which are feeding this live packet gatherer".format(
                    provenance_data[1]))))

        return provenance_items

    @overrides(AbstractHasAssociatedBinary.get_binary_file_name)
    def get_binary_file_name(self):
        return 'live_packet_gather.aplx'

    @overrides(AbstractHasAssociatedBinary.get_binary_start_type)
    def get_binary_start_type(self):
        return ExecutableType.USES_SIMULATION_INTERFACE

    @inject_items({
        "machine_time_step": "MachineTimeStep",
        "time_scale_factor": "TimeScaleFactor",
        "tags": "MemoryTags"})
    @overrides(
        AbstractGeneratesDataSpecification.generate_data_specification,
        additional_arguments={
            "machine_time_step", "time_scale_factor", "tags"
        })
    def generate_data_specification(
            self, spec, placement,  # @UnusedVariable
            machine_time_step, time_scale_factor, tags):
        """
        :param int machine_time_step:
        :param int time_scale_factor:
        :param ~pacman.model.tags.Tags tags:
        """
        # pylint: disable=too-many-arguments, arguments-differ
        spec.comment("\n*** Spec for LivePacketGather Instance ***\n\n")

        # Construct the data images needed for the Neuron:
        self._reserve_memory_regions(spec)
        self._write_setup_info(spec, machine_time_step, time_scale_factor)
        self._write_configuration_region(
            spec, tags.get_ip_tags_for_vertex(self))

        # End-of-Spec:
        spec.end_specification()

    def _reserve_memory_regions(self, spec):
        """ Reserve SDRAM space for memory areas.

        :param ~.DataSpecificationGenerator spec:
        """
        spec.comment("\nReserving memory space for data regions:\n\n")

        # Reserve memory:
        spec.reserve_memory_region(
            region=self._REGIONS.SYSTEM,
            size=SIMULATION_N_BYTES, label='system')
        spec.reserve_memory_region(
            region=self._REGIONS.CONFIG,
            size=self._CONFIG_SIZE, label='config')
        self.reserve_provenance_data_region(spec)

    def _write_configuration_region(self, spec, iptags):
        """ Write the configuration region to the spec

        :param ~.DataSpecificationGenerator spec:
        :param iterable(~.IPTag) iptags:
            The set of IP tags assigned to the object
        :raise ConfigurationException: if `iptags` is empty
        :raise DataSpecificationException:
            when something goes wrong with the DSG generation
        """
        spec.switch_write_focus(region=self._REGIONS.CONFIG)

        spec.write_value(int(self._lpg_params.use_prefix))
        spec.write_value(self._lpg_params.key_prefix or 0)
        spec.write_value(self._lpg_params.prefix_type.value
                         if self._lpg_params.prefix_type else 0)
        spec.write_value(self._lpg_params.message_type.value
                         if self._lpg_params.message_type else 0)
        spec.write_value(self._lpg_params.right_shift)
        spec.write_value(int(self._lpg_params.payload_as_time_stamps))
        spec.write_value(int(self._lpg_params.use_payload_prefix))
        spec.write_value(self._lpg_params.payload_prefix or 0)
        spec.write_value(data=self._lpg_params.payload_right_shift)

        # SDP tag
        for iptag in iptags:
            spec.write_value(iptag.tag)
            spec.write_value(_ONE_SHORT.unpack(_TWO_BYTES.pack(
                iptag.destination_y, iptag.destination_x))[0])
            break
        else:
            raise ConfigurationException("no iptag provided")

        # number of packets to send per time stamp
        spec.write_value(self._lpg_params.number_of_packets_sent_per_time_step)

    def _write_setup_info(self, spec, machine_time_step, time_scale_factor):
        """ Write basic info to the system region

        :param ~.DataSpecificationGenerator spec:
        :param int machine_time_step:
        :param int time_scale_factor:
        """
        # Write this to the system region (to be picked up by the simulation):
        spec.switch_write_focus(region=self._REGIONS.SYSTEM)
        spec.write_array(get_simulation_header_array(
            self.get_binary_file_name(), machine_time_step, time_scale_factor))

    @staticmethod
    def get_cpu_usage():
        """ Get the CPU used by this vertex

        :return: 0
        :rtype: int
        """
        return 0

    @classmethod
    def get_sdram_usage(cls):
        """ Get the SDRAM used by this vertex

        :rtype: int
        """
        return (
            SYSTEM_BYTES_REQUIREMENT + cls._CONFIG_SIZE +
            cls.get_provenance_data_size(cls._N_ADDITIONAL_PROVENANCE_ITEMS))

    @classmethod
    def get_dtcm_usage(cls):
        """ Get the DTCM used by this vertex

        :rtype: int
        """
        return cls._CONFIG_SIZE



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import itertools
import json
import string
from spinn_front_end_common.utilities.helpful_functions import (
    generate_unique_folder_name)

_VALID_CHARS = frozenset(
    "-_.() {}{}".format(string.ascii_letters, string.digits))


class ProvenanceJSONWriter(object):
    """ Write provenance data into JSON
    """

    __slots__ = []

    def __call__(self, provenance_data_items, provenance_data_path):
        """
        :param list(ProvenanceDataItem) provenance_data_items:
        :param str provenance_data_path:
        """
        # Group data by the first name
        items = sorted(provenance_data_items, key=lambda item: item.names[0])
        for name, group in itertools.groupby(
                items, lambda item: item.names[0]):
            # Create a root node
            root = dict()

            # Go through the items and add them
            for item in group:
                # Add the "categories" for the item (any name between the first
                # and last)
                parent = self._build_path(root, item)
                # Add the item
                parent[item.names[-1]] = str(item.value)

            # write json form into file provided
            with open(self._get_file(provenance_data_path, name), "w") as f:
                json.dump(root, f, indent=4, separators=(',', ': '))

    @staticmethod
    def _build_path(root, item):
        """
        :param dict(str,object) root:
        :param ProvenanceDataItem item:
        :rtype: dict(str,object)
        """
        parent = root
        for name in item.names[1:-1]:
            if not (name in parent and isinstance(parent[name], dict)):
                # If there isn't already a category of this name under the
                # super element, create a new category under the parent
                parent[name] = dict()
            parent = parent[name]
        return parent

    @staticmethod
    def _get_file(path, name):
        """
        :param str path:
        :param str name:
        :rtype: str
        """
        remapped = "".join(c if c in _VALID_CHARS else '_' for c in name)
        return generate_unique_folder_name(path, remapped, ".json")



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinn_utilities.overrides import overrides
from pacman.model.partitioner_interfaces import LegacyPartitionerAPI
from pacman.model.graphs.application import ApplicationVertex
from pacman.model.resources import (
    ConstantSDRAM, CPUCyclesPerTickResource, DTCMResource, ResourceContainer)
from .live_packet_gather_machine_vertex import LivePacketGatherMachineVertex


class LivePacketGather(ApplicationVertex, LegacyPartitionerAPI):
    """ A model which stores all the events it receives during a timer tick\
        and then compresses them into Ethernet packets and sends them out of\
        a SpiNNaker machine.
    """

    def __init__(self, lpg_params, constraints=None):
        """
        :param LivePacketGatherParameters lpg_params:
        :param constraints:
        :type constraints:
            iterable(~pacman.model.constraints.AbstractConstraint)
        """
        label = lpg_params.label or "Live Packet Gatherer"
        super().__init__(label, constraints, 1)
        self._lpg_params = lpg_params

    @overrides(LegacyPartitionerAPI.create_machine_vertex)
    def create_machine_vertex(
            self, vertex_slice, resources_required,
            label=None, constraints=None):
        machine_vertex = LivePacketGatherMachineVertex(
            self._lpg_params, constraints, self, label)
        if vertex_slice:
            assert (vertex_slice == machine_vertex.vertex_slice)
        if resources_required:
            assert (resources_required == machine_vertex.resources_required)
        return machine_vertex

    @property
    @overrides(LegacyPartitionerAPI.n_atoms)
    def n_atoms(self):
        return 1

    @overrides(LegacyPartitionerAPI.get_resources_used_by_atoms)
    def get_resources_used_by_atoms(self, vertex_slice):  # @UnusedVariable
        return ResourceContainer(
            sdram=ConstantSDRAM(
                LivePacketGatherMachineVertex.get_sdram_usage()),
            dtcm=DTCMResource(LivePacketGatherMachineVertex.get_dtcm_usage()),
            cpu_cycles=CPUCyclesPerTickResource(
                LivePacketGatherMachineVertex.get_cpu_usage()),
            iptags=[self._lpg_params.get_iptag_resource()])



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
from spinn_utilities.progress_bar import ProgressBar
from spinn_utilities.log import FormatAdapter
from spinn_front_end_common.utilities.utility_objs import ProvenanceDataItem

logger = FormatAdapter(logging.getLogger(__name__))


class RouterProvenanceGatherer(object):
    """ Gathers diagnostics from the routers.
    """

    __slots__ = [
        # int for how many packets were sent
        '_total_sent_packets',

        # how many new packets were received
        '_total_new_packets',

        # how many dropped packets
        '_total_dropped_packets',

        # total missed dropped packets
        '_total_missed_dropped_packets',

        # total lost dropped packets
        '_total_lost_dropped_packets',

        # machine
        '_machine',
        # placements
        '_placements',
        # transceiver
        '_txrx',
    ]

    def __call__(
            self, transceiver, machine, router_tables, using_reinjection,
            provenance_data_objects=None, extra_monitor_vertices=None,
            placements=None):
        """
        :param ~spinnman.transceiver.Transceiver transceiver:
            the SpiNNMan interface object
        :param ~spinn_machine.Machine machine:
            the SpiNNaker machine
        :param router_tables: the router tables that have been generated
        :type router_tables:
            ~pacman.model.routing_tables.MulticastRoutingTables
        :param bool using_reinjection: whether we are reinjecting packets
        :param list(ProvenanceDataItem) provenance_data_objects:
            any existing provenance information to add to
        :param list(ExtraMonitorSupportMachineVertex) extra_monitor_vertices:
            vertices which represent the extra monitor code
        :param ~pacman.model.placements.Placements placements:
            the placements object
        """
        # pylint: disable=too-many-arguments
        # pylint: disable=attribute-defined-outside-init
        self._total_sent_packets = 0
        self._total_new_packets = 0
        self._total_dropped_packets = 0
        self._total_missed_dropped_packets = 0
        self._total_lost_dropped_packets = 0
        self._txrx = transceiver
        self._machine = machine
        self._placements = placements

        if provenance_data_objects is not None:
            prov_items = provenance_data_objects
        else:
            prov_items = list()

        self._add_router_provenance_data(
            router_tables, extra_monitor_vertices, prov_items)

        prov_items.append(ProvenanceDataItem(
            ["router_provenance", "total_multi_cast_sent_packets"],
            self._total_sent_packets))
        prov_items.append(ProvenanceDataItem(
            ["router_provenance", "total_created_packets"],
            self._total_new_packets))
        prov_items.append(ProvenanceDataItem(
            ["router_provenance", "total_dropped_packets"],
            self._total_dropped_packets))
        prov_items.append(ProvenanceDataItem(
            ["router_provenance", "total_missed_dropped_packets"],
            self._total_missed_dropped_packets))
        prov_items.append(ProvenanceDataItem(
            ["router_provenance", "total_lost_dropped_packets"],
            self._total_lost_dropped_packets))

        return prov_items

    def _add_router_provenance_data(
            self, router_tables, extra_monitor_vertices, items):
        """ Writes the provenance data of the router diagnostics

        :param ~.MulticastRoutingTables router_tables:
            the routing tables generated by PACMAN
        :param list(ExtraMonitorSupportMachineVertex) extra_monitor_vertices:
            list of extra monitor vertices
        :param list(ProvenanceDataItem) items:
        """
        # pylint: disable=too-many-arguments
        progress = ProgressBar(self._machine.n_chips*2,
                               "Getting Router Provenance")

        seen_chips = set()

        # get all extra monitor core data if it exists
        reinjection_data = None
        if extra_monitor_vertices is not None:
            monitor = extra_monitor_vertices[0]
            reinjection_data = monitor.get_reinjection_status_for_vertices(
                placements=self._placements,
                extra_monitor_cores_for_data=extra_monitor_vertices,
                transceiver=self._txrx)

        for router_table in progress.over(sorted(
                router_tables.routing_tables,
                key=lambda table: (table.x, table.y)), False):
            self._add_router_table_diagnostic(
                router_table.x, router_table.y, seen_chips,
                router_table, items, reinjection_data)

        for chip in progress.over(sorted(
                self._machine.chips, key=lambda c: (c.x, c.y))):
            self._add_router_chip_diagnostic(
                chip, seen_chips, items, reinjection_data)

    def _add_router_table_diagnostic(
            self, x, y, seen_chips, router_table, items, reinjection_data):
        """
        :param int x:
        :param int y:
        :param set(tuple(int,int)) seen_chips:
        :param ~.MulticastRoutingTable router_table:
        :param list(ProvenanceDataItem) items:
        :param dict(tuple(int,int),ReInjectionStatus) reinjection_data:
        """
        # pylint: disable=too-many-arguments, bare-except
        if not self._machine.get_chip_at(x, y).virtual:
            try:
                router_diagnostic = self._txrx.get_router_diagnostics(x, y)
                seen_chips.add((x, y))
                reinjection_status = None
                if reinjection_data is not None:
                    reinjection_status = reinjection_data[x, y]
                self.__add_router_diagnostics(
                    items, x, y, router_diagnostic, reinjection_status, True,
                    router_table)
                self._add_totals(router_diagnostic, reinjection_status)
            except:  # noqa: E722
                logger.warning(
                    "Could not read routing diagnostics from {}, {}",
                    x, y, exc_info=True)

    def _add_router_chip_diagnostic(
            self, chip, seen_chips, items, reinjection_data):
        """
        :param ~.Chip chip:
        :param set(tuple(int,int)) seen_chips:
        :param list(ProvenanceDataItem) items:
        :param dict(tuple(int,int),ReInjectionStatus) reinjection_data:
        """
        # pylint: disable=bare-except
        if not chip.virtual and (chip.x, chip.y) not in seen_chips:
            try:
                diagnostic = self._txrx.get_router_diagnostics(chip.x, chip.y)

                if (diagnostic.n_dropped_multicast_packets or
                        diagnostic.n_local_multicast_packets or
                        diagnostic.n_external_multicast_packets):
                    reinjection_status = None
                    if reinjection_data is not None:
                        reinjection_status = reinjection_data[chip.x, chip.y]
                    self.__add_router_diagnostics(
                        items, chip.x, chip.y, diagnostic, reinjection_status,
                        False, None)
                    self._add_totals(diagnostic, reinjection_status)
            except:  # noqa: E722
                # There could be issues with unused chips - don't worry!
                pass

    def _add_totals(self, router_diagnostic, reinjection_status):
        """
        :param ~.RouterDiagnostics router_diagnostic:
        :param ReInjectionStatus reinjection_status:
        """
        self._total_sent_packets += (
            router_diagnostic.n_local_multicast_packets +
            router_diagnostic.n_external_multicast_packets)
        self._total_new_packets += router_diagnostic.n_local_multicast_packets
        self._total_dropped_packets += (
            router_diagnostic.n_dropped_multicast_packets)
        if reinjection_status is not None:
            self._total_missed_dropped_packets += (
                reinjection_status.n_missed_dropped_packets)
            self._total_lost_dropped_packets += (
                reinjection_status.n_dropped_packet_overflows)
        else:
            self._total_lost_dropped_packets += (
                router_diagnostic.n_dropped_multicast_packets)

    @staticmethod
    def __add_name(names, name):
        """
        :param list(str) names:
        :param str name:
        :rtype: list(str)
        """
        new_names = list(names)
        new_names.append(name)
        return new_names

    def __add_router_diagnostics(
            self, items, x, y, router_diagnostic, reinjection_status,
            expected, router_table):
        """ Stores router diagnostics as a set of provenance data items.

        :param list(ProvenanceDataItem) items: the list to append to
        :param int x: x coordinate of the router in question
        :param int y: y coordinate of the router in question
        :param ~.RouterDiagnostics router_diagnostic:
            the router diagnostic object
        :param ReInjectionStatus reinjection_status:
            the data gained from the extra monitor re-injection subsystem
        :param ~.MulticastRoutingTable router_table:
            the router table generated by the PACMAN tools
        """
        # pylint: disable=too-many-arguments
        names = list()
        names.append("router_provenance")
        if expected:
            names.append("expected_routers")
        else:
            names.append("unexpected_routers")
        names.append("router_at_chip_{}_{}".format(x, y))

        items.append(ProvenanceDataItem(
            self.__add_name(names, "Local_Multicast_Packets"),
            str(router_diagnostic.n_local_multicast_packets)))
        items.append(ProvenanceDataItem(
            self.__add_name(names, "External_Multicast_Packets"),
            str(router_diagnostic.n_external_multicast_packets)))

        # simplify the if by making components of it outside.
        has_dropped = router_diagnostic.n_dropped_multicast_packets > 0
        missing_stuff = None
        has_reinjection = reinjection_status is not None
        if has_reinjection:
            missing_stuff = (
                (reinjection_status.n_dropped_packets +
                 reinjection_status.n_missed_dropped_packets +
                 reinjection_status.n_dropped_packet_overflows +
                 reinjection_status.n_reinjected_packets +
                 reinjection_status.n_link_dumps +
                 reinjection_status.n_processor_dumps) <
                router_diagnostic.n_dropped_multicast_packets)
        items.append(ProvenanceDataItem(
            self.__add_name(names, "Dropped_Multicast_Packets"),
            str(router_diagnostic.n_dropped_multicast_packets),
            report=((has_dropped and not has_reinjection) or (
                has_dropped and has_reinjection and missing_stuff)),
            message=(
                "The router on {}, {} has dropped {} multicast route packets. "
                "Try increasing the machine_time_step and/or the time scale "
                "factor or reducing the number of atoms per core."
                .format(x, y, router_diagnostic.n_dropped_multicast_packets))))
        items.append(ProvenanceDataItem(
            self.__add_name(
                names, "Dropped_Multicast_Packets_via_local_transmission"),
            str(router_diagnostic.user_3),
            report=(router_diagnostic.user_3 > 0),
            message=(
                "The router on {}, {} has dropped {} multicast packets that"
                " were transmitted by local cores. This occurs where the "
                "router has no entry associated with the multi-cast key. "
                "Try investigating the keys allocated to the vertices and "
                "the router table entries for this chip.".format(
                    x, y, router_diagnostic.user_3))))
        items.append(ProvenanceDataItem(
            self.__add_name(
                names, "default_routed_external_multicast_packets"),
            str(router_diagnostic.user_2),
            report=(router_diagnostic.user_2 > 0 and
                    ((router_table is not None and
                      router_table.number_of_defaultable_entries == 0) or
                     router_table is None)),
            message=(
                "The router on {}, {} has default routed {} multicast packets,"
                " but the router table did not expect any default routed "
                "packets. This occurs where the router has no entry"
                " associated with the multi-cast key. "
                "Try investigating the keys allocated to the vertices and "
                "the router table entries for this chip.".format(
                    x, y, router_diagnostic.user_2))))

        items.append(ProvenanceDataItem(
            self.__add_name(names, "Local_P2P_Packets"),
            str(router_diagnostic.n_local_peer_to_peer_packets)))
        items.append(ProvenanceDataItem(
            self.__add_name(names, "External_P2P_Packets"),
            str(router_diagnostic.n_external_peer_to_peer_packets)))
        items.append(ProvenanceDataItem(
            self.__add_name(names, "Dropped_P2P_Packets"),
            str(router_diagnostic.n_dropped_peer_to_peer_packets)))
        items.append(ProvenanceDataItem(
            self.__add_name(names, "Local_NN_Packets"),
            str(router_diagnostic.n_local_nearest_neighbour_packets)))
        items.append(ProvenanceDataItem(
            self.__add_name(names, "External_NN_Packets"),
            str(router_diagnostic.n_external_nearest_neighbour_packets)))
        items.append(ProvenanceDataItem(
            self.__add_name(names, "Dropped_NN_Packets"),
            str(router_diagnostic.n_dropped_nearest_neighbour_packets)))
        items.append(ProvenanceDataItem(
            self.__add_name(names, "Local_FR_Packets"),
            str(router_diagnostic.n_local_fixed_route_packets)))
        items.append(ProvenanceDataItem(
            self.__add_name(names, "External_FR_Packets"),
            str(router_diagnostic.n_external_fixed_route_packets)))
        items.append(ProvenanceDataItem(
            self.__add_name(names, "Dropped_FR_Packets"),
            str(router_diagnostic.n_dropped_fixed_route_packets),
            report=router_diagnostic.n_dropped_fixed_route_packets > 0,
            message=(
                "The router on chip {}:{} dropped {} Fixed route packets. "
                "This is indicative of a error within the data extraction "
                "process as this is the only expected user of fixed route "
                "packets.".format(
                    x, y, router_diagnostic.n_dropped_fixed_route_packets))))
        if reinjection_status is not None:
            items.append(ProvenanceDataItem(
                self.__add_name(names, "Received_For_Reinjection"),
                reinjection_status.n_dropped_packets))
            items.append(ProvenanceDataItem(
                self.__add_name(names, "Missed_For_Reinjection"),
                reinjection_status.n_missed_dropped_packets,
                report=reinjection_status.n_missed_dropped_packets > 0,
                message=(
                    "The extra monitor on {}, {} has missed {} "
                    "packets.".format(
                        x, y, reinjection_status.n_missed_dropped_packets))))
            items.append(ProvenanceDataItem(
                self.__add_name(names, "Reinjection_Overflows"),
                reinjection_status.n_dropped_packet_overflows,
                report=reinjection_status.n_dropped_packet_overflows > 0,
                message=(
                    "The extra monitor on {}, {} has dropped {} "
                    "packets.".format(
                        x, y,
                        reinjection_status.n_dropped_packet_overflows))))
            items.append(ProvenanceDataItem(
                self.__add_name(names, "Reinjected"),
                reinjection_status.n_reinjected_packets))
            items.append(ProvenanceDataItem(
                self.__add_name(names, "Dumped_from_a_Link"),
                str(reinjection_status.n_link_dumps),
                report=(reinjection_status.n_link_dumps > 0 and
                        self._has_virtual_chip_connected(self._machine, x, y)),
                message=(
                    "The extra monitor on {}, {} has detected that {} packets "
                    "were dumped from a outgoing link of this chip's router."
                    " This often occurs when external devices are used in the "
                    "script but not connected to the communication fabric "
                    "correctly. These packets may have been reinjected "
                    "multiple times and so this number may be a overestimate."
                    .format(x, y, reinjection_status.n_link_dumps))))
            items.append(ProvenanceDataItem(
                self.__add_name(names, "Dumped_from_a_processor"),
                str(reinjection_status.n_processor_dumps),
                report=reinjection_status.n_processor_dumps > 0,
                message=(
                    "The extra monitor on {}, {} has detected that {} packets "
                    "were dumped from a core failing to take the packet."
                    " This often occurs when the executable has crashed or"
                    " has not been given a multicast packet callback. It can"
                    " also result from the core taking too long to process"
                    " each packet. These packets were reinjected and so this"
                    " number is likely a overestimate.".format(
                        x, y, reinjection_status.n_processor_dumps))))

        items.append(ProvenanceDataItem(
            self.__add_name(names, "Error status"),
            str(router_diagnostic.error_status),
            report=router_diagnostic.error_status > 0,
            message=(
                "The router on {}, {} has a non-zero error status.  This could"
                " indicate a hardware fault.  The errors set are {}, and the"
                " error count is {}".format(
                    x, y, router_diagnostic.errors_set,
                    router_diagnostic.error_count))))

    @staticmethod
    def _has_virtual_chip_connected(machine, x, y):
        """
        :param ~.Machine machine:
        :param int x:
        :param int y:
        :rtype: bool
        """
        for link in machine.get_chip_at(x, y).router.links:
            if machine.get_chip_at(
                    link.destination_x, link.destination_y).virtual:
                return True
        return False



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import os
from pacman.model.routing_tables.compressed_multicast_routing_table import (
    CompressedMulticastRoutingTable)
from spinn_utilities.progress_bar import ProgressBar
from pacman.operations.algorithm_reports import reports

_FOLDER_NAME = "routing_tables_from_machine"


class RoutingTableFromMachineReport(object):
    """ Report the routing table that was actually on the machine.
    """

    def __call__(
            self, report_default_directory, routing_tables):
        """
        :param str report_default_directory:
        :param routing_tables:
        :type routing_tables:
            ~pacman.model.routing_tables.MulticastRoutingTables
        :param ~spinnman.transceiver.Transceiver transceiver:
        :param int app_id:
        """
        # pylint: disable=protected-access
        tables = list(routing_tables.routing_tables)
        progress = ProgressBar(tables, "Reading Routing Tables from Machine")

        folder_name = os.path.join(report_default_directory, _FOLDER_NAME)
        os.mkdir(folder_name)

        # generate a file for every multicast entry
        for routing_table in progress.over(tables):
            reports.generate_routing_table(routing_table, folder_name)

    @staticmethod
    def _read_routing_table(txrx, table, app_id):
        machine_routing_table = \
            CompressedMulticastRoutingTable(table.x, table.y)
        for routing_entry in txrx.get_multicast_routes(
                table.x, table.y, app_id):
            machine_routing_table.add_multicast_routing_entry(routing_entry)
        return machine_routing_table



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .read_status_process import ReadStatusProcess
from .reset_counters_process import ResetCountersProcess
from .set_packet_types_process import SetPacketTypesProcess
from .set_router_timeout_process import SetRouterTimeoutProcess
from .clear_queue_process import ClearQueueProcess
from .load_application_mc_routes_process import LoadApplicationMCRoutesProcess
from .load_system_mc_routes_process import LoadSystemMCRoutesProcess

__all__ = [
    "ClearQueueProcess",
    "LoadApplicationMCRoutesProcess",
    "LoadSystemMCRoutesProcess",
    "ReadStatusProcess",
    "ResetCountersProcess",
    "SetPacketTypesProcess",
    "SetRouterTimeoutProcess",
    ]
